"""
GitLab API:
https://docs.gitlab.com/ce/api/keys.html
"""

import base64
import hashlib


def key_fingerprint(key: str) -> str:
    key_part = key.split()[1]
    decoded = base64.b64decode(key_part.encode("ascii"))
    digest = hashlib.sha256(decoded).digest()
    return f"SHA256:{base64.b64encode(digest).rstrip(b'=').decode('utf-8')}"


def test_keys_ssh(gl, user, SSH_KEY):
    key = user.keys.create({"title": "foo@bar", "key": SSH_KEY})

    # Get key by ID (admin only).
    key_by_id = gl.keys.get(key.id)
    assert key_by_id.title == key.title
    assert key_by_id.key == key.key

    fingerprint = key_fingerprint(SSH_KEY)
    # Get key by fingerprint (admin only).
    key_by_fingerprint = gl.keys.get(fingerprint=fingerprint)
    assert key_by_fingerprint.title == key.title
    assert key_by_fingerprint.key == key.key

    key.delete()


def test_keys_deploy(gl, project, DEPLOY_KEY):
    key = project.keys.create({"title": "foo@bar", "key": DEPLOY_KEY})

    fingerprint = key_fingerprint(DEPLOY_KEY)
    key_by_fingerprint = gl.keys.get(fingerprint=fingerprint)
    assert key_by_fingerprint.title == key.title
    assert key_by_fingerprint.key == key.key
    assert len(key_by_fingerprint.deploy_keys_projects) == 1

    key.delete()



import pickle

import pytest

import gitlab
from gitlab import base
from tests.unit import helpers
from tests.unit.helpers import FakeManager  # noqa: F401, needed for _create_managers


def test_instantiate(gl, fake_manager):
    attrs = {"foo": "bar"}
    obj = helpers.FakeObject(fake_manager, attrs.copy())

    assert attrs == obj._attrs
    assert {} == obj._updated_attrs
    assert obj._create_managers() is None
    assert fake_manager == obj.manager
    assert gl == obj.manager.gitlab
    assert str(obj) == f"{type(obj)} => {attrs}"


def test_instantiate_non_dict(gl, fake_manager):
    with pytest.raises(gitlab.exceptions.GitlabParsingError):
        helpers.FakeObject(fake_manager, ["a", "list", "fails"])


def test_missing_attribute_does_not_raise_custom(gl, fake_manager):
    """Ensure a missing attribute does not raise our custom error message
    if the RESTObject was not created from a list"""
    obj = helpers.FakeObject(manager=fake_manager, attrs={"foo": "bar"})
    with pytest.raises(AttributeError) as excinfo:
        obj.missing_attribute
    exc_str = str(excinfo.value)
    assert "missing_attribute" in exc_str
    assert "was created via a list()" not in exc_str
    assert base._URL_ATTRIBUTE_ERROR not in exc_str


def test_missing_attribute_from_list_raises_custom(gl, fake_manager):
    """Ensure a missing attribute raises our custom error message if the
    RESTObject was created from a list"""
    obj = helpers.FakeObject(
        manager=fake_manager, attrs={"foo": "bar"}, created_from_list=True
    )
    with pytest.raises(AttributeError) as excinfo:
        obj.missing_attribute
    exc_str = str(excinfo.value)
    assert "missing_attribute" in exc_str
    assert "was created via a list()" in exc_str
    assert base._URL_ATTRIBUTE_ERROR in exc_str


def test_picklability(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "bar"})
    original_obj_module = obj._module
    pickled = pickle.dumps(obj)
    unpickled = pickle.loads(pickled)
    assert isinstance(unpickled, helpers.FakeObject)
    assert hasattr(unpickled, "_module")
    assert unpickled._module == original_obj_module
    pickle.dumps(unpickled)


def test_attrs(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "bar"})

    assert "bar" == obj.foo
    with pytest.raises(AttributeError):
        getattr(obj, "bar")

    obj.bar = "baz"
    assert "baz" == obj.bar
    assert {"foo": "bar"} == obj._attrs
    assert {"bar": "baz"} == obj._updated_attrs


def test_get_id(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "bar"})
    obj.id = 42
    assert 42 == obj.get_id()

    obj.id = None
    assert obj.get_id() is None


def test_encoded_id(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "bar"})
    obj.id = 42
    assert 42 == obj.encoded_id

    obj.id = None
    assert obj.encoded_id is None

    obj.id = "plain"
    assert "plain" == obj.encoded_id

    obj.id = "a/path"
    assert "a%2Fpath" == obj.encoded_id

    # If you assign it again it does not double URL-encode
    obj.id = obj.encoded_id
    assert "a%2Fpath" == obj.encoded_id


def test_custom_id_attr(fake_manager):
    obj = helpers.OtherFakeObject(fake_manager, {"foo": "bar"})
    assert "bar" == obj.get_id()


def test_update_attrs(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "bar"})
    obj.bar = "baz"
    obj._update_attrs({"foo": "foo", "bar": "bar"})
    assert {"foo": "foo", "bar": "bar"} == obj._attrs
    assert {} == obj._updated_attrs


def test_update_attrs_deleted(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"foo": "foo", "bar": "bar"})
    obj.bar = "baz"
    obj._update_attrs({"foo": "foo"})
    assert {"foo": "foo"} == obj._attrs
    assert {} == obj._updated_attrs


def test_dir_unique(fake_manager):
    obj = helpers.FakeObject(fake_manager, {"manager": "foo"})
    assert len(dir(obj)) == len(set(dir(obj)))


def test_create_managers(gl, fake_manager):
    class ObjectWithManager(helpers.FakeObject):
        fakes: "FakeManager"

    obj = ObjectWithManager(fake_manager, {"foo": "bar"})
    obj.id = 42
    assert isinstance(obj.fakes, helpers.FakeManager)
    assert obj.fakes.gitlab == gl
    assert obj.fakes._parent == obj


def test_equality(fake_manager):
    obj1 = helpers.FakeObject(fake_manager, {"id": "foo"})
    obj2 = helpers.FakeObject(fake_manager, {"id": "foo", "other_attr": "bar"})
    assert obj1 == obj2
    assert len(set((obj1, obj2))) == 1


def test_equality_custom_id(fake_manager):
    obj1 = helpers.OtherFakeObject(fake_manager, {"foo": "bar"})
    obj2 = helpers.OtherFakeObject(fake_manager, {"foo": "bar", "other_attr": "baz"})
    assert obj1 == obj2


def test_equality_no_id(fake_manager):
    obj1 = helpers.FakeObject(fake_manager, {"attr1": "foo"})
    obj2 = helpers.FakeObject(fake_manager, {"attr1": "bar"})
    assert not obj1 == obj2


def test_inequality(fake_manager):
    obj1 = helpers.FakeObject(fake_manager, {"id": "foo"})
    obj2 = helpers.FakeObject(fake_manager, {"id": "bar"})
    assert obj1 != obj2


def test_inequality_no_id(fake_manager):
    obj1 = helpers.FakeObject(fake_manager, {"attr1": "foo"})
    obj2 = helpers.FakeObject(fake_manager, {"attr1": "bar"})
    assert obj1 != obj2
    assert len(set((obj1, obj2))) == 2


def test_equality_with_other_objects(fake_manager):
    obj1 = helpers.FakeObject(fake_manager, {"id": "foo"})
    obj2 = None
    assert not obj1 == obj2


def test_dunder_str(fake_manager):
    fake_object = helpers.FakeObject(fake_manager, {"attr1": "foo"})
    assert str(fake_object) == (
        "<class 'tests.unit.helpers.FakeObject'> => {'attr1': 'foo'}"
    )


@pytest.mark.parametrize(
    "id_attr,repr_attr, attrs, expected_repr",
    [
        ("id", None, {"id": 1}, "<ReprObject id:1>"),
        (
            "id",
            "name",
            {"id": 1, "name": "fake"},
            "<ReprObject id:1 name:fake>",
        ),
        ("name", "name", {"name": "fake"}, "<ReprObject name:fake>"),
        ("id", "name", {"id": 1}, "<ReprObject id:1>"),
        (None, None, {}, "<ReprObject>"),
        (None, "name", {"name": "fake"}, "<ReprObject name:fake>"),
        (None, "name", {}, "<ReprObject>"),
    ],
    ids=[
        "GetMixin with id",
        "GetMixin with id and _repr_attr",
        "GetMixin with _repr_attr matching _id_attr",
        "GetMixin with _repr_attr without _repr_attr value defined",
        "GetWithoutIDMixin",
        "GetWithoutIDMixin with _repr_attr",
        "GetWithoutIDMixin with _repr_attr without _repr_attr value defined",
    ],
)
def test_dunder_repr(fake_manager, id_attr, repr_attr, attrs, expected_repr):
    class ReprObject(helpers.FakeObject):
        _id_attr = id_attr
        _repr_attr = repr_attr

    fake_object = ReprObject(fake_manager, attrs)

    assert repr(fake_object) == expected_repr


def test_pformat(fake_manager):
    fake_object = helpers.FakeObject(
        fake_manager, {"attr1": "foo" * 10, "ham": "eggs" * 15}
    )
    assert fake_object.pformat() == (
        "<class 'tests.unit.helpers.FakeObject'> => "
        "\n{'attr1': 'foofoofoofoofoofoofoofoofoofoo',\n"
        " 'ham': 'eggseggseggseggseggseggseggseggseggseggseggseggseggseggseggs'}"
    )


def test_pprint(capfd, fake_manager):
    fake_object = helpers.FakeObject(
        fake_manager, {"attr1": "foo" * 10, "ham": "eggs" * 15}
    )
    result = fake_object.pprint()
    assert result is None
    stdout, stderr = capfd.readouterr()
    assert stdout == (
        "<class 'tests.unit.helpers.FakeObject'> => "
        "\n{'attr1': 'foofoofoofoofoofoofoofoofoofoo',\n"
        " 'ham': 'eggseggseggseggseggseggseggseggseggseggseggseggseggseggseggs'}\n"
    )
    assert stderr == ""


def test_repr(fake_manager):
    attrs = {"attr1": "foo"}
    obj = helpers.FakeObject(fake_manager, attrs)
    assert repr(obj) == "<FakeObject id:None>"

    helpers.FakeObject._id_attr = None
    assert repr(obj) == "<FakeObject>"


def test_attributes_get(fake_object):
    assert fake_object.attr1 == "foo"
    result = fake_object.attributes
    assert result == {"attr1": "foo", "alist": [1, 2, 3]}


def test_attributes_shows_updates(fake_object):
    # Updated attribute value is reflected in `attributes`
    fake_object.attr1 = "hello"
    assert fake_object.attributes == {"attr1": "hello", "alist": [1, 2, 3]}
    assert fake_object.attr1 == "hello"
    # New attribute is in `attributes`
    fake_object.new_attrib = "spam"
    assert fake_object.attributes == {
        "attr1": "hello",
        "new_attrib": "spam",
        "alist": [1, 2, 3],
    }


def test_attributes_is_copy(fake_object):
    # Modifying the dictionary does not cause modifications to the object
    result = fake_object.attributes
    result["alist"].append(10)
    assert result == {"attr1": "foo", "alist": [1, 2, 3, 10]}
    assert fake_object.attributes == {"attr1": "foo", "alist": [1, 2, 3]}


def test_attributes_has_parent_attrs(fake_object_with_parent):
    assert fake_object_with_parent.attr1 == "foo"
    result = fake_object_with_parent.attributes
    assert result == {"attr1": "foo", "alist": [1, 2, 3], "test_id": "42"}


def test_asdict(fake_object):
    assert fake_object.attr1 == "foo"
    result = fake_object.asdict()
    assert result == {"attr1": "foo", "alist": [1, 2, 3]}


def test_asdict_no_parent_attrs(fake_object_with_parent):
    assert fake_object_with_parent.attr1 == "foo"
    result = fake_object_with_parent.asdict()
    assert result == {"attr1": "foo", "alist": [1, 2, 3]}
    assert "test_id" not in fake_object_with_parent.asdict()
    assert "test_id" not in fake_object_with_parent.asdict(with_parent_attrs=False)
    assert "test_id" in fake_object_with_parent.asdict(with_parent_attrs=True)


def test_asdict_modify_dict_does_not_change_object(fake_object):
    result = fake_object.asdict()
    # Demonstrate modifying the dictionary does not modify the object
    result["attr1"] = "testing"
    result["alist"].append(4)
    assert result == {"attr1": "testing", "alist": [1, 2, 3, 4]}
    assert fake_object.attr1 == "foo"
    assert fake_object.alist == [1, 2, 3]


def test_asdict_modify_dict_does_not_change_object2(fake_object):
    # Modify attribute and then ensure modifying a list in the returned dict won't
    # modify the list in the object.
    fake_object.attr1 = [9, 7, 8]
    assert fake_object.asdict() == {
        "attr1": [9, 7, 8],
        "alist": [1, 2, 3],
    }
    result = fake_object.asdict()
    result["attr1"].append(1)
    assert fake_object.asdict() == {
        "attr1": [9, 7, 8],
        "alist": [1, 2, 3],
    }


def test_asdict_modify_object(fake_object):
    # asdict() returns the updated value
    fake_object.attr1 = "spam"
    assert fake_object.asdict() == {"attr1": "spam", "alist": [1, 2, 3]}



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import NoUpdateMixin, ObjectDeleteMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectTag",
    "ProjectTagManager",
    "ProjectProtectedTag",
    "ProjectProtectedTagManager",
]


class ProjectTag(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class ProjectTagManager(NoUpdateMixin, RESTManager):
    _path = "/projects/{project_id}/repository/tags"
    _obj_cls = ProjectTag
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("tag_name", "ref"), optional=("message",)
    )

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> ProjectTag:
        return cast(ProjectTag, super().get(id=id, lazy=lazy, **kwargs))


class ProjectProtectedTag(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class ProjectProtectedTagManager(NoUpdateMixin, RESTManager):
    _path = "/projects/{project_id}/protected_tags"
    _obj_cls = ProjectProtectedTag
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name",), optional=("create_access_level",)
    )

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectProtectedTag:
        return cast(ProjectProtectedTag, super().get(id=id, lazy=lazy, **kwargs))



def test_current_user_email(gl):
    gl.auth()
    mail = gl.user.emails.create({"email": "current@user.com"})
    assert mail in gl.user.emails.list()

    mail.delete()


def test_current_user_gpg_keys(gl, GPG_KEY):
    gl.auth()
    gkey = gl.user.gpgkeys.create({"key": GPG_KEY})
    assert gkey in gl.user.gpgkeys.list()

    # Seems broken on the gitlab side
    gkey = gl.user.gpgkeys.get(gkey.id)

    gkey.delete()


def test_current_user_ssh_keys(gl, SSH_KEY):
    gl.auth()
    key = gl.user.keys.create({"title": "testkey", "key": SSH_KEY})
    assert key in gl.user.keys.list()

    key.delete()


def test_current_user_status(gl):
    gl.auth()
    message = "Test"
    emoji = "thumbsup"
    status = gl.user.status.get()

    status.message = message
    status.emoji = emoji
    status.save()

    new_status = gl.user.status.get()
    assert new_status.message == message
    assert new_status.emoji == emoji



"""
GitLab API: https://docs.gitlab.com/ee/api/container_registry.html
"""

import re

import pytest
import responses

from gitlab.v4.objects import ProjectRegistryRepository, RegistryRepository

repositories_content = [
    {
        "id": 1,
        "name": "",
        "path": "group/project",
        "project_id": 9,
        "location": "gitlab.example.com:5000/group/project",
        "created_at": "2019-01-10T13:38:57.391Z",
        "cleanup_policy_started_at": "2020-01-10T15:40:57.391Z",
    },
    {
        "id": 2,
        "name": "releases",
        "path": "group/project/releases",
        "project_id": 9,
        "location": "gitlab.example.com:5000/group/project/releases",
        "created_at": "2019-01-10T13:39:08.229Z",
        "cleanup_policy_started_at": "2020-08-17T03:12:35.489Z",
    },
]


@pytest.fixture
def resp_list_registry_repositories():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/(groups|projects)/1/registry/repositories"
            ),
            json=repositories_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_registry_repository():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/registry/repositories/1",
            json=repositories_content[0],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_registry_repository():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/registry/repositories/1",
            status=204,
        )
        yield rsps


def test_list_group_registry_repositories(group, resp_list_registry_repositories):
    repositories = group.registry_repositories.list()
    assert isinstance(repositories[0], ProjectRegistryRepository)
    assert repositories[0].id == 1


def test_list_project_registry_repositories(project, resp_list_registry_repositories):
    repositories = project.repositories.list()
    assert isinstance(repositories[0], ProjectRegistryRepository)
    assert repositories[0].id == 1


def test_delete_project_registry_repository(project, resp_delete_registry_repository):
    project.repositories.delete(1)


def test_get_registry_repository(gl, resp_get_registry_repository):
    repository = gl.registry_repositories.get(1)
    assert isinstance(repository, RegistryRepository)
    assert repository.id == 1



def test_project_boards(project):
    assert not project.boards.list()

    board = project.boards.create({"name": "testboard"})
    board = project.boards.get(board.id)

    project.boards.delete(board.id)


def test_group_boards(group):
    assert not group.boards.list()

    board = group.boards.create({"name": "testboard"})
    board = group.boards.get(board.id)

    group.boards.delete(board.id)



import gitlab


def test_create_issue(project):
    issue = project.issues.create({"title": "my issue 1"})
    issue2 = project.issues.create({"title": "my issue 2"})

    issues = project.issues.list()
    issue_iids = [issue.iid for issue in issues]
    assert {issue, issue2} <= set(issues)

    # Test 'iids' as a list
    filtered_issues = project.issues.list(iids=issue_iids)
    assert {issue, issue2} == set(filtered_issues)

    issue2.state_event = "close"
    issue2.save()
    assert issue in project.issues.list(state="opened")
    assert issue2 in project.issues.list(state="closed")

    participants = issue.participants()
    assert participants
    assert isinstance(participants, list)
    assert type(issue.closed_by()) == list
    assert type(issue.related_merge_requests()) == list


def test_issue_notes(issue):
    note = issue.notes.create({"body": "This is an issue note"})
    assert note in issue.notes.list()

    emoji = note.awardemojis.create({"name": "tractor"})
    assert emoji in note.awardemojis.list()

    emoji.delete()
    note.delete()


def test_issue_labels(project, issue):
    project.labels.create({"name": "label2", "color": "#aabbcc"})
    issue.labels = ["label2"]
    issue.save()

    assert issue in project.issues.list(labels=["label2"])
    assert issue in project.issues.list(labels="label2")
    assert issue in project.issues.list(labels="Any")
    assert issue not in project.issues.list(labels="None")


def test_issue_links(project, issue):
    linked_issue = project.issues.create({"title": "Linked issue"})
    source_issue, target_issue = issue.links.create(
        {"target_project_id": project.id, "target_issue_iid": linked_issue.iid}
    )
    assert source_issue == issue
    assert target_issue == linked_issue

    links = issue.links.list()
    assert links

    link_id = links[0].issue_link_id

    issue.links.delete(link_id)


def test_issue_label_events(issue):
    events = issue.resourcelabelevents.list()
    assert isinstance(events, list)

    event = issue.resourcelabelevents.get(events[0].id)
    assert isinstance(event, gitlab.v4.objects.ProjectIssueResourceLabelEvent)


def test_issue_weight_events(issue):
    issue.weight = 13
    issue.save()

    events = issue.resource_weight_events.list()
    assert isinstance(events, list)

    event = issue.resource_weight_events.get(events[0].id)
    assert isinstance(event, gitlab.v4.objects.ProjectIssueResourceWeightEvent)


def test_issue_milestones(project, milestone):
    data = {"title": "my issue 1", "milestone_id": milestone.id}
    issue = project.issues.create(data)
    assert milestone.issues().next().title == "my issue 1"

    milestone_events = issue.resourcemilestoneevents.list()
    assert isinstance(milestone_events, list)

    milestone_event = issue.resourcemilestoneevents.get(milestone_events[0].id)
    assert isinstance(
        milestone_event, gitlab.v4.objects.ProjectIssueResourceMilestoneEvent
    )

    assert issue in project.issues.list(milestone=milestone.title)


def test_issue_discussions(issue):
    discussion = issue.discussions.create({"body": "Discussion body"})
    assert discussion in issue.discussions.list()

    d_note = discussion.notes.create({"body": "first note"})
    d_note_from_get = discussion.notes.get(d_note.id)
    d_note_from_get.body = "updated body"
    d_note_from_get.save()

    discussion = issue.discussions.get(discussion.id)
    assert discussion.attributes["notes"][-1]["body"] == "updated body"

    d_note_from_get.delete()



import pytest

import gitlab


@pytest.mark.gitlab_premium
def test_project_push_rules(project):
    with pytest.raises(gitlab.GitlabParsingError):
        # when no rules are defined the API call returns back `None` which
        # causes a gitlab.GitlabParsingError in RESTObject.__init__()
        project.pushrules.get()

    push_rules = project.pushrules.create({"deny_delete_tag": True})
    assert push_rules.deny_delete_tag

    push_rules.deny_delete_tag = False
    push_rules.save()

    push_rules = project.pushrules.get()
    assert push_rules
    assert not push_rules.deny_delete_tag

    push_rules.delete()

    with pytest.raises(gitlab.GitlabParsingError):
        project.pushrules.get()



"""
GitLab API: https://docs.gitlab.com/ce/api/keys.html
"""

import pytest
import responses

from gitlab.v4.objects import Key

key_content = {"id": 1, "title": "title", "key": "ssh-keytype AAAAC3Nza/key comment"}


@pytest.fixture
def resp_get_key_by_id():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/keys/1",
            json=key_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_key_by_fingerprint():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/keys?fingerprint=foo",
            json=key_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_key_by_id(gl, resp_get_key_by_id):
    key = gl.keys.get(1)
    assert isinstance(key, Key)
    assert key.id == 1
    assert key.title == "title"


def test_get_key_by_fingerprint(gl, resp_get_key_by_fingerprint):
    key = gl.keys.get(fingerprint="foo")
    assert isinstance(key, Key)
    assert key.id == 1
    assert key.title == "title"


def test_get_key_missing_attrs(gl):
    with pytest.raises(AttributeError):
        gl.keys.get()



"""
Ensure objects defined in gitlab.v4.objects have REST* as last item in class
definition

Original notes by John L. Villalovos

An example of an incorrect definition:
    class ProjectPipeline(RESTObject, RefreshMixin, ObjectDeleteMixin):
                          ^^^^^^^^^^ This should be at the end.

Correct way would be:
    class ProjectPipeline(RefreshMixin, ObjectDeleteMixin, RESTObject):
                                      Correctly at the end ^^^^^^^^^^


Why this is an issue:

  When we do type-checking for gitlab/mixins.py we make RESTObject or
  RESTManager the base class for the mixins

  Here is how our classes look when type-checking:

      class RESTObject:
          def __init__(self, manager: "RESTManager", attrs: Dict[str, Any]) -> None:
              ...

      class Mixin(RESTObject):
          ...

      # Wrong ordering here
      class Wrongv4Object(RESTObject, RefreshMixin):
          ...

  If we actually ran this in Python we would get the following error:
         class Wrongv4Object(RESTObject, Mixin):
    TypeError: Cannot create a consistent method resolution
    order (MRO) for bases RESTObject, Mixin

  When we are type-checking it fails to understand the class Wrongv4Object
  and thus we can't type check it correctly.

Almost all classes in gitlab/v4/objects/*py were already correct before this
check was added.
"""

import inspect

import pytest

import gitlab.v4.objects


def test_show_issue() -> None:
    """Test case to demonstrate the TypeError that occurs"""

    class RESTObject:
        def __init__(self, manager: str, attrs: int) -> None: ...

    class Mixin(RESTObject): ...

    with pytest.raises(TypeError) as exc_info:
        # Wrong ordering here
        class Wrongv4Object(RESTObject, Mixin):  # type: ignore
            ...

    # The error message in the exception should be:
    #   TypeError: Cannot create a consistent method resolution
    #   order (MRO) for bases RESTObject, Mixin

    # Make sure the exception string contains "MRO"
    assert "MRO" in exc_info.exconly()

    # Correctly ordered class, no exception
    class Correctv4Object(Mixin, RESTObject): ...


def test_mros() -> None:
    """Ensure objects defined in gitlab.v4.objects have REST* as last item in
    class definition.

    We do this as we need to ensure the MRO (Method Resolution Order) is
    correct.
    """

    failed_messages = []
    for module_name, module_value in inspect.getmembers(gitlab.v4.objects):
        if not inspect.ismodule(module_value):
            # We only care about the modules
            continue
        # Iterate through all the classes in our module
        for class_name, class_value in inspect.getmembers(module_value):
            if not inspect.isclass(class_value):
                continue

            # Ignore imported classes from gitlab.base
            if class_value.__module__ == "gitlab.base":
                continue

            mro = class_value.mro()

            # We only check classes which have a 'gitlab.base' class in their MRO
            has_base = False
            for count, obj in enumerate(mro, start=1):
                if obj.__module__ == "gitlab.base":
                    has_base = True
                    base_classname = obj.__name__
            if has_base:
                filename = inspect.getfile(class_value)
                # NOTE(jlvillal): The very last item 'mro[-1]' is always going
                # to be 'object'. That is why we are checking 'mro[-2]'.
                if mro[-2].__module__ != "gitlab.base":
                    failed_messages.append(
                        (
                            f"class definition for {class_name!r} in file {filename!r} "
                            f"must have {base_classname!r} as the last class in the "
                            f"class definition"
                        )
                    )
    failed_msg = "\n".join(failed_messages)
    assert not failed_messages, failed_msg



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupBoardList",
    "GroupBoardListManager",
    "GroupBoard",
    "GroupBoardManager",
    "ProjectBoardList",
    "ProjectBoardListManager",
    "ProjectBoard",
    "ProjectBoardManager",
]


class GroupBoardList(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupBoardListManager(CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/boards/{board_id}/lists"
    _obj_cls = GroupBoardList
    _from_parent_attrs = {"group_id": "group_id", "board_id": "id"}
    _create_attrs = RequiredOptional(
        exclusive=("label_id", "assignee_id", "milestone_id")
    )
    _update_attrs = RequiredOptional(required=("position",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> GroupBoardList:
        return cast(GroupBoardList, super().get(id=id, lazy=lazy, **kwargs))


class GroupBoard(SaveMixin, ObjectDeleteMixin, RESTObject):
    lists: GroupBoardListManager


class GroupBoardManager(CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/boards"
    _obj_cls = GroupBoard
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> GroupBoard:
        return cast(GroupBoard, super().get(id=id, lazy=lazy, **kwargs))


class ProjectBoardList(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectBoardListManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/boards/{board_id}/lists"
    _obj_cls = ProjectBoardList
    _from_parent_attrs = {"project_id": "project_id", "board_id": "id"}
    _create_attrs = RequiredOptional(
        exclusive=("label_id", "assignee_id", "milestone_id")
    )
    _update_attrs = RequiredOptional(required=("position",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectBoardList:
        return cast(ProjectBoardList, super().get(id=id, lazy=lazy, **kwargs))


class ProjectBoard(SaveMixin, ObjectDeleteMixin, RESTObject):
    lists: ProjectBoardListManager


class ProjectBoardManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/boards"
    _obj_cls = ProjectBoard
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectBoard:
        return cast(ProjectBoard, super().get(id=id, lazy=lazy, **kwargs))



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, NoUpdateMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "Hook",
    "HookManager",
    "ProjectHook",
    "ProjectHookManager",
    "GroupHook",
    "GroupHookManager",
]


class Hook(ObjectDeleteMixin, RESTObject):
    _url = "/hooks"
    _repr_attr = "url"


class HookManager(NoUpdateMixin, RESTManager):
    _path = "/hooks"
    _obj_cls = Hook
    _create_attrs = RequiredOptional(required=("url",))

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> Hook:
        return cast(Hook, super().get(id=id, lazy=lazy, **kwargs))


class ProjectHook(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "url"


class ProjectHookManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/hooks"
    _obj_cls = ProjectHook
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "enable_ssl_verification",
            "token",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "job_events",
            "pipeline_events",
            "wiki_events",
            "enable_ssl_verification",
            "token",
        ),
    )

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectHook:
        return cast(ProjectHook, super().get(id=id, lazy=lazy, **kwargs))


class GroupHook(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "url"


class GroupHookManager(CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/hooks"
    _obj_cls = GroupHook
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "confidential_note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "deployment_events",
            "releases_events",
            "subgroup_events",
            "enable_ssl_verification",
            "token",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "confidential_note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "deployment_events",
            "releases_events",
            "subgroup_events",
            "enable_ssl_verification",
            "token",
        ),
    )

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> GroupHook:
        return cast(GroupHook, super().get(id=id, lazy=lazy, **kwargs))



"""
GitLab API: https://docs.gitlab.com/ee/api/cluster_agents.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectClusterAgent

agent_content = {
    "id": 1,
    "name": "agent-1",
    "config_project": {
        "id": 20,
        "description": "",
        "name": "test",
        "name_with_namespace": "Administrator / test",
        "path": "test",
        "path_with_namespace": "root/test",
        "created_at": "2022-03-20T20:42:40.221Z",
    },
    "created_at": "2022-04-20T20:42:40.221Z",
    "created_by_user_id": 42,
}


@pytest.fixture
def resp_list_project_cluster_agents():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/cluster_agents",
            json=[agent_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_project_cluster_agent():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/cluster_agents/1",
            json=agent_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_project_cluster_agent():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/cluster_agents",
            json=agent_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_delete_project_cluster_agent():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/cluster_agents/1",
            status=204,
        )
        yield rsps


def test_list_project_cluster_agents(project, resp_list_project_cluster_agents):
    agent = project.cluster_agents.list()[0]
    assert isinstance(agent, ProjectClusterAgent)
    assert agent.name == "agent-1"


def test_get_project_cluster_agent(project, resp_get_project_cluster_agent):
    agent = project.cluster_agents.get(1)
    assert isinstance(agent, ProjectClusterAgent)
    assert agent.name == "agent-1"


def test_create_project_cluster_agent(project, resp_create_project_cluster_agent):
    agent = project.cluster_agents.create({"name": "agent-1"})
    assert isinstance(agent, ProjectClusterAgent)
    assert agent.name == "agent-1"


def test_delete_project_cluster_agent(project, resp_delete_project_cluster_agent):
    agent = project.cluster_agents.get(1, lazy=True)
    agent.delete()



from gitlab.mixins import (
    AccessRequestMixin,
    SetMixin,
    SubscribableMixin,
    TimeTrackingMixin,
    TodoMixin,
    UserAgentDetailMixin,
)


def test_access_request_mixin():
    class TestClass(AccessRequestMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "approve")


def test_subscribable_mixin():
    class TestClass(SubscribableMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "subscribe")
    assert hasattr(obj, "unsubscribe")


def test_todo_mixin():
    class TestClass(TodoMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "todo")


def test_time_tracking_mixin():
    class TestClass(TimeTrackingMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "time_stats")
    assert hasattr(obj, "time_estimate")
    assert hasattr(obj, "reset_time_estimate")
    assert hasattr(obj, "add_spent_time")
    assert hasattr(obj, "reset_spent_time")


def test_set_mixin():
    class TestClass(SetMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "set")


def test_user_agent_detail_mixin():
    class TestClass(UserAgentDetailMixin):
        pass

    obj = TestClass()
    assert hasattr(obj, "user_agent_detail")



"""
GitLab API:
https://docs.gitlab.com/ee/api/repositories.html
https://docs.gitlab.com/ee/api/repository_files.html
"""

from urllib.parse import quote

import pytest
import responses

from gitlab.v4.objects import ProjectFile

file_path = "app/models/key.rb"
ref = "main"


@pytest.fixture
def resp_get_repository_file():
    file_response = {
        "file_name": "key.rb",
        "file_path": file_path,
        "size": 1476,
        "encoding": "base64",
        "content": "IyA9PSBTY2hlbWEgSW5mb3...",
        "content_sha256": "4c294617b60715c1d218e61164a3abd4808a4284cbc30e6728a01ad9aada4481",
        "ref": ref,
        "blob_id": "79f7bbd25901e8334750839545a9bd021f0e4c83",
        "commit_id": "d5a3ff139356ce33e37e73add446f16869741b50",
        "last_commit_id": "570e7b2abdd848b95f2f578043fc23bd6f6fd24d",
    }

    encoded_path = quote(file_path, safe="")

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=f"http://localhost/api/v4/projects/1/repository/files/{encoded_path}",
            json=file_response,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_repository_file(project, resp_get_repository_file):
    file = project.files.get(file_path, ref=ref)
    assert isinstance(file, ProjectFile)
    assert file.file_path == file_path



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ListMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "PagesDomain",
    "PagesDomainManager",
    "ProjectPagesDomain",
    "ProjectPagesDomainManager",
]


class PagesDomain(RESTObject):
    _id_attr = "domain"


class PagesDomainManager(ListMixin, RESTManager):
    _path = "/pages/domains"
    _obj_cls = PagesDomain


class ProjectPagesDomain(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "domain"


class ProjectPagesDomainManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/pages/domains"
    _obj_cls = ProjectPagesDomain
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("domain",), optional=("certificate", "key")
    )
    _update_attrs = RequiredOptional(optional=("certificate", "key"))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectPagesDomain:
        return cast(ProjectPagesDomain, super().get(id=id, lazy=lazy, **kwargs))



def test_project_deploy_keys(gl, project, DEPLOY_KEY):
    deploy_key = project.keys.create({"title": "foo@bar", "key": DEPLOY_KEY})
    assert deploy_key in project.keys.list()

    project2 = gl.projects.create({"name": "deploy-key-project"})
    project2.keys.enable(deploy_key.id)
    assert deploy_key in project2.keys.list()

    project2.keys.delete(deploy_key.id)

    project2.delete()



import pytest
import responses

from gitlab import exceptions

ci_lint_create_content = {"status": "valid", "errors": [], "warnings": []}
ci_lint_create_invalid_content = {
    "status": "invalid",
    "errors": ["invalid format"],
    "warnings": [],
}


project_ci_lint_content = {
    "valid": True,
    "merged_yaml": "---\n:test_job:\n  :script: echo 1\n",
    "errors": [],
    "warnings": [],
}


@pytest.fixture
def resp_create_ci_lint():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/ci/lint",
            json=ci_lint_create_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_ci_lint_invalid():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/ci/lint",
            json=ci_lint_create_invalid_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_project_ci_lint():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/ci/lint",
            json=project_ci_lint_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_project_ci_lint():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/ci/lint",
            json=project_ci_lint_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_ci_lint_create(gl, resp_create_ci_lint, valid_gitlab_ci_yml):
    lint_result = gl.ci_lint.create({"content": valid_gitlab_ci_yml})
    assert lint_result.status == "valid"


def test_ci_lint_validate(gl, resp_create_ci_lint, valid_gitlab_ci_yml):
    gl.ci_lint.validate({"content": valid_gitlab_ci_yml})


def test_ci_lint_validate_invalid_raises(
    gl, resp_create_ci_lint_invalid, invalid_gitlab_ci_yml
):
    with pytest.raises(exceptions.GitlabCiLintError, match="invalid format"):
        gl.ci_lint.validate({"content": invalid_gitlab_ci_yml})


def test_project_ci_lint_get(project, resp_get_project_ci_lint):
    lint_result = project.ci_lint.get()
    assert lint_result.valid is True


def test_project_ci_lint_create(
    project, resp_create_project_ci_lint, valid_gitlab_ci_yml
):
    lint_result = project.ci_lint.create({"content": valid_gitlab_ci_yml})
    assert lint_result.valid is True



from typing import Any, cast, Union

from gitlab import types
from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "DeployToken",
    "DeployTokenManager",
    "GroupDeployToken",
    "GroupDeployTokenManager",
    "ProjectDeployToken",
    "ProjectDeployTokenManager",
]


class DeployToken(ObjectDeleteMixin, RESTObject):
    pass


class DeployTokenManager(ListMixin, RESTManager):
    _path = "/deploy_tokens"
    _obj_cls = DeployToken


class GroupDeployToken(ObjectDeleteMixin, RESTObject):
    pass


class GroupDeployTokenManager(RetrieveMixin, CreateMixin, DeleteMixin, RESTManager):
    _path = "/groups/{group_id}/deploy_tokens"
    _from_parent_attrs = {"group_id": "id"}
    _obj_cls = GroupDeployToken
    _create_attrs = RequiredOptional(
        required=(
            "name",
            "scopes",
        ),
        optional=(
            "expires_at",
            "username",
        ),
    )
    _list_filters = ("scopes",)
    _types = {"scopes": types.ArrayAttribute}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> GroupDeployToken:
        return cast(GroupDeployToken, super().get(id=id, lazy=lazy, **kwargs))


class ProjectDeployToken(ObjectDeleteMixin, RESTObject):
    pass


class ProjectDeployTokenManager(RetrieveMixin, CreateMixin, DeleteMixin, RESTManager):
    _path = "/projects/{project_id}/deploy_tokens"
    _from_parent_attrs = {"project_id": "id"}
    _obj_cls = ProjectDeployToken
    _create_attrs = RequiredOptional(
        required=(
            "name",
            "scopes",
        ),
        optional=(
            "expires_at",
            "username",
        ),
    )
    _list_filters = ("scopes",)
    _types = {"scopes": types.ArrayAttribute}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectDeployToken:
        return cast(ProjectDeployToken, super().get(id=id, lazy=lazy, **kwargs))



"""
GitLab API:
https://docs.gitlab.com/ee/api/resource_groups.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectResourceGroup, ProjectResourceGroupUpcomingJob

from .test_jobs import failed_job_content

resource_group_content = {
    "id": 3,
    "key": "production",
    "process_mode": "unordered",
    "created_at": "2021-09-01T08:04:59.650Z",
    "updated_at": "2021-09-01T08:04:59.650Z",
}


@pytest.fixture
def resp_list_resource_groups():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/resource_groups",
            json=[resource_group_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_resource_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/resource_groups/production",
            json=resource_group_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_upcoming_jobs():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/resource_groups/production/upcoming_jobs",
            json=[failed_job_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_resource_groups(project, resp_list_resource_groups):
    resource_groups = project.resource_groups.list()
    assert isinstance(resource_groups, list)
    assert isinstance(resource_groups[0], ProjectResourceGroup)
    assert resource_groups[0].process_mode == "unordered"


def test_get_project_resource_group(project, resp_get_resource_group):
    resource_group = project.resource_groups.get("production")
    assert isinstance(resource_group, ProjectResourceGroup)
    assert resource_group.process_mode == "unordered"


def test_list_resource_group_upcoming_jobs(project, resp_list_upcoming_jobs):
    resource_group = project.resource_groups.get("production", lazy=True)
    upcoming_jobs = resource_group.upcoming_jobs.list()

    assert isinstance(upcoming_jobs, list)
    assert isinstance(upcoming_jobs[0], ProjectResourceGroupUpcomingJob)
    assert upcoming_jobs[0].ref == "main"



import copy

import pytest
import responses

from gitlab.const import DEFAULT_URL


def test_list_instance_variables(gitlab_cli, gl):
    cmd = ["variable", "list"]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_variables(gitlab_cli, group):
    cmd = ["group-variable", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_project_variables(gitlab_cli, project):
    cmd = ["project-variable", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_project_variables_with_path(gitlab_cli, project):
    cmd = ["project-variable", "list", "--project-id", project.path_with_namespace]
    ret = gitlab_cli(cmd)

    assert ret.success


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_list_project_variables_with_path_url_check(script_runner, resp_get_project):
    resp_get_project_variables = copy.deepcopy(resp_get_project)
    resp_get_project_variables.update(
        url=f"{DEFAULT_URL}/api/v4/projects/project%2Fwith%2Fa%2Fnamespace/variables"
    )
    resp_get_project_variables.update(json=[])

    responses.add(**resp_get_project_variables)
    ret = script_runner.run(
        [
            "gitlab",
            "project-variable",
            "list",
            "--project-id",
            "project/with/a/namespace",
        ]
    )
    assert ret.success



from unittest.mock import mock_open, patch

import pytest
import requests
import responses

from gitlab import base, GitlabUploadError
from gitlab import types as gl_types
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    GetMixin,
    GetWithoutIdMixin,
    ListMixin,
    RefreshMixin,
    SaveMixin,
    SetMixin,
    UpdateMethod,
    UpdateMixin,
    UploadMixin,
)


class FakeObject(base.RESTObject):
    pass


class FakeManager(base.RESTManager):
    _path = "/tests"
    _obj_cls = FakeObject


@responses.activate
def test_get_mixin(gl):
    class M(GetMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.GET,
        url=url,
        json={"id": 42, "foo": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = mgr.get(42)
    assert isinstance(obj, FakeObject)
    assert obj.foo == "bar"
    assert obj.id == 42
    assert obj._lazy is False
    assert responses.assert_call_count(url, 1) is True


def test_get_mixin_lazy(gl):
    class M(GetMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests/42"

    mgr = M(gl)
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url=url,
            json={"id": 42, "foo": "bar"},
            status=200,
            match=[responses.matchers.query_param_matcher({})],
        )
        obj = mgr.get(42, lazy=True)
    assert isinstance(obj, FakeObject)
    assert not hasattr(obj, "foo")
    assert obj.id == 42
    assert obj._lazy is True
    # a `lazy` get does not make a network request
    assert not rsps.calls


def test_get_mixin_lazy_missing_attribute(gl):
    class FakeGetManager(GetMixin, FakeManager):
        pass

    manager = FakeGetManager(gl)
    obj = manager.get(1, lazy=True)
    assert obj.id == 1
    with pytest.raises(AttributeError) as exc:
        obj.missing_attribute
    # undo `textwrap.fill()`
    message = str(exc.value).replace("\n", " ")
    assert "'FakeObject' object has no attribute 'missing_attribute'" in message
    assert (
        "note that <class 'tests.unit.mixins.test_mixin_methods.FakeObject'> was "
        "created as a `lazy` object and was not initialized with any data."
    ) in message


@responses.activate
def test_head_mixin(gl):
    class M(GetMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.HEAD,
        url=url,
        headers={"X-GitLab-Header": "test"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    manager = M(gl)
    result = manager.head(42)
    assert isinstance(result, requests.structures.CaseInsensitiveDict)
    assert result["x-gitlab-header"] == "test"


@responses.activate
def test_refresh_mixin(gl):
    class TestClass(RefreshMixin, FakeObject):
        pass

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.GET,
        url=url,
        json={"id": 42, "foo": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = FakeManager(gl)
    obj = TestClass(mgr, {"id": 42})
    res = obj.refresh()
    assert res is None
    assert obj.foo == "bar"
    assert obj.id == 42
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_get_without_id_mixin(gl):
    class M(GetWithoutIdMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests"
    responses.add(
        method=responses.GET,
        url=url,
        json={"foo": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = mgr.get()
    assert isinstance(obj, FakeObject)
    assert obj.foo == "bar"
    assert not hasattr(obj, "id")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_list_mixin(gl):
    class M(ListMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests"
    headers = {
        "X-Page": "1",
        "X-Next-Page": "2",
        "X-Per-Page": "1",
        "X-Total-Pages": "2",
        "X-Total": "2",
        "Link": ("<http://localhost/api/v4/tests" ' rel="next"'),
    }
    responses.add(
        method=responses.GET,
        headers=headers,
        url=url,
        json=[{"id": 42, "foo": "bar"}, {"id": 43, "foo": "baz"}],
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    # test RESTObjectList
    mgr = M(gl)
    obj_list = mgr.list(iterator=True)
    assert isinstance(obj_list, base.RESTObjectList)
    assert obj_list.current_page == 1
    assert obj_list.prev_page is None
    assert obj_list.next_page == 2
    assert obj_list.per_page == 1
    assert obj_list.total == 2
    assert obj_list.total_pages == 2
    assert len(obj_list) == 2

    for obj in obj_list:
        assert isinstance(obj, FakeObject)
        assert obj.id in (42, 43)

    # test list()
    obj_list = mgr.list(get_all=True)
    assert isinstance(obj_list, list)
    assert obj_list[0].id == 42
    assert obj_list[1].id == 43
    assert isinstance(obj_list[0], FakeObject)
    assert len(obj_list) == 2
    assert responses.assert_call_count(url, 2) is True


@responses.activate
def test_list_mixin_with_attributes(gl):
    class M(ListMixin, FakeManager):
        _types = {"my_array": gl_types.ArrayAttribute}

    url = "http://localhost/api/v4/tests"
    responses.add(
        method=responses.GET,
        headers={},
        url=url,
        json=[],
        status=200,
        match=[responses.matchers.query_param_matcher({"my_array[]": ["1", "2", "3"]})],
    )

    mgr = M(gl)
    mgr.list(iterator=True, my_array=[1, 2, 3])


@responses.activate
def test_list_other_url(gl):
    class M(ListMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/others"
    responses.add(
        method=responses.GET,
        url=url,
        json=[{"id": 42, "foo": "bar"}],
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj_list = mgr.list(path="/others", iterator=True)
    assert isinstance(obj_list, base.RESTObjectList)
    obj = obj_list.next()
    assert obj.id == 42
    assert obj.foo == "bar"
    with pytest.raises(StopIteration):
        obj_list.next()


def test_create_mixin_missing_attrs(gl):
    class M(CreateMixin, FakeManager):
        _create_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )

    mgr = M(gl)
    data = {"foo": "bar", "baz": "blah"}
    mgr._create_attrs.validate_attrs(data=data)

    data = {"baz": "blah"}
    with pytest.raises(AttributeError) as error:
        mgr._create_attrs.validate_attrs(data=data)
    assert "foo" in str(error.value)


@responses.activate
def test_create_mixin(gl):
    class M(CreateMixin, FakeManager):
        _create_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )
        _update_attrs = gl_types.RequiredOptional(required=("foo",), optional=("bam",))

    url = "http://localhost/api/v4/tests"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "foo": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = mgr.create({"foo": "bar"})
    assert isinstance(obj, FakeObject)
    assert obj.id == 42
    assert obj.foo == "bar"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_create_mixin_custom_path(gl):
    class M(CreateMixin, FakeManager):
        _create_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )
        _update_attrs = gl_types.RequiredOptional(required=("foo",), optional=("bam",))

    url = "http://localhost/api/v4/others"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "foo": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = mgr.create({"foo": "bar"}, path="/others")
    assert isinstance(obj, FakeObject)
    assert obj.id == 42
    assert obj.foo == "bar"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_create_mixin_with_attributes(gl):
    class M(CreateMixin, FakeManager):
        _types = {"my_array": gl_types.ArrayAttribute}

    url = "http://localhost/api/v4/tests"
    responses.add(
        method=responses.POST,
        headers={},
        url=url,
        json={},
        status=200,
        match=[responses.matchers.json_params_matcher({"my_array": [1, 2, 3]})],
    )

    mgr = M(gl)
    mgr.create({"my_array": [1, 2, 3]})


def test_update_mixin_missing_attrs(gl):
    class M(UpdateMixin, FakeManager):
        _update_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )

    mgr = M(gl)
    data = {"foo": "bar", "baz": "blah"}
    mgr._update_attrs.validate_attrs(data=data)

    data = {"baz": "blah"}
    with pytest.raises(AttributeError) as error:
        mgr._update_attrs.validate_attrs(data=data)
    assert "foo" in str(error.value)


@responses.activate
def test_update_mixin(gl):
    class M(UpdateMixin, FakeManager):
        _create_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )
        _update_attrs = gl_types.RequiredOptional(required=("foo",), optional=("bam",))

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.PUT,
        url=url,
        json={"id": 42, "foo": "baz"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    server_data = mgr.update(42, {"foo": "baz"})
    assert isinstance(server_data, dict)
    assert server_data["id"] == 42
    assert server_data["foo"] == "baz"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_update_mixin_uses_post(gl):
    class M(UpdateMixin, FakeManager):
        _update_method = UpdateMethod.POST

    url = "http://localhost/api/v4/tests/1"
    responses.add(
        method=responses.POST,
        url=url,
        json={},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    mgr.update(1, {})
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_update_mixin_no_id(gl):
    class M(UpdateMixin, FakeManager):
        _create_attrs = gl_types.RequiredOptional(
            required=("foo",), optional=("bar", "baz")
        )
        _update_attrs = gl_types.RequiredOptional(required=("foo",), optional=("bam",))

    url = "http://localhost/api/v4/tests"
    responses.add(
        method=responses.PUT,
        url=url,
        json={"foo": "baz"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    server_data = mgr.update(new_data={"foo": "baz"})
    assert isinstance(server_data, dict)
    assert server_data["foo"] == "baz"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_delete_mixin(gl):
    class M(DeleteMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.DELETE,
        url=url,
        json="",
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    mgr.delete(42)
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_save_mixin(gl):
    class M(UpdateMixin, FakeManager):
        pass

    class TestClass(SaveMixin, base.RESTObject):
        pass

    url = "http://localhost/api/v4/tests/42"
    responses.add(
        method=responses.PUT,
        url=url,
        json={"id": 42, "foo": "baz"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = TestClass(mgr, {"id": 42, "foo": "bar"})
    obj.foo = "baz"
    obj.save()
    assert obj._attrs["foo"] == "baz"
    assert obj._updated_attrs == {}
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_save_mixin_without_new_data(gl):
    class M(UpdateMixin, FakeManager):
        pass

    class TestClass(SaveMixin, base.RESTObject):
        pass

    url = "http://localhost/api/v4/tests/1"
    responses.add(method=responses.PUT, url=url)

    mgr = M(gl)
    obj = TestClass(mgr, {"id": 1, "foo": "bar"})
    obj.save()

    assert obj._attrs["foo"] == "bar"
    assert responses.assert_call_count(url, 0) is True


@responses.activate
def test_set_mixin(gl):
    class M(SetMixin, FakeManager):
        pass

    url = "http://localhost/api/v4/tests/foo"
    responses.add(
        method=responses.PUT,
        url=url,
        json={"key": "foo", "value": "bar"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = M(gl)
    obj = mgr.set("foo", "bar")
    assert isinstance(obj, FakeObject)
    assert obj.key == "foo"
    assert obj.value == "bar"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_upload_mixin_with_filepath_and_filedata(gl):
    class TestClass(UploadMixin, FakeObject):
        _upload_path = "/tests/{id}/uploads"

    url = "http://localhost/api/v4/tests/42/uploads"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "file_name": "test.txt", "file_content": "testing contents"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = FakeManager(gl)
    obj = TestClass(mgr, {"id": 42})
    with pytest.raises(
        GitlabUploadError, match="File contents and file path specified"
    ):
        obj.upload("test.txt", "testing contents", "/home/test.txt")


@responses.activate
def test_upload_mixin_without_filepath_nor_filedata(gl):
    class TestClass(UploadMixin, FakeObject):
        _upload_path = "/tests/{id}/uploads"

    url = "http://localhost/api/v4/tests/42/uploads"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "file_name": "test.txt", "file_content": "testing contents"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = FakeManager(gl)
    obj = TestClass(mgr, {"id": 42})
    with pytest.raises(GitlabUploadError, match="No file contents or path specified"):
        obj.upload("test.txt")


@responses.activate
def test_upload_mixin_with_filedata(gl):
    class TestClass(UploadMixin, FakeObject):
        _upload_path = "/tests/{id}/uploads"

    url = "http://localhost/api/v4/tests/42/uploads"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "file_name": "test.txt", "file_content": "testing contents"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = FakeManager(gl)
    obj = TestClass(mgr, {"id": 42})
    res_only_data = obj.upload("test.txt", "testing contents")
    assert obj._get_upload_path() == "/tests/42/uploads"
    assert isinstance(res_only_data, dict)
    assert res_only_data["file_name"] == "test.txt"
    assert res_only_data["file_content"] == "testing contents"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_upload_mixin_with_filepath(gl):
    class TestClass(UploadMixin, FakeObject):
        _upload_path = "/tests/{id}/uploads"

    url = "http://localhost/api/v4/tests/42/uploads"
    responses.add(
        method=responses.POST,
        url=url,
        json={"id": 42, "file_name": "test.txt", "file_content": "testing contents"},
        status=200,
        match=[responses.matchers.query_param_matcher({})],
    )

    mgr = FakeManager(gl)
    obj = TestClass(mgr, {"id": 42})
    with patch("builtins.open", mock_open(read_data="raw\nfile\ndata")):
        res_only_path = obj.upload("test.txt", None, "/filepath")
    assert obj._get_upload_path() == "/tests/42/uploads"
    assert isinstance(res_only_path, dict)
    assert res_only_path["file_name"] == "test.txt"
    assert res_only_path["file_content"] == "testing contents"
    assert responses.assert_call_count(url, 1) is True



import dataclasses
import datetime
import logging
import pathlib
import tempfile
import time
import uuid
from subprocess import check_output
from typing import Optional

import pytest
import requests

import gitlab
import gitlab.base
from tests.functional import helpers
from tests.functional.fixtures.docker import *  # noqa

SLEEP_TIME = 10


@dataclasses.dataclass
class GitlabVersion:
    major: int
    minor: int
    patch: str
    revision: str

    def __post_init__(self):
        self.major, self.minor = int(self.major), int(self.minor)


@pytest.fixture(scope="session")
def gitlab_version(gl) -> GitlabVersion:
    version, revision = gl.version()
    major, minor, patch = version.split(".")
    return GitlabVersion(major=major, minor=minor, patch=patch, revision=revision)


@pytest.fixture(scope="session")
def fixture_dir(test_dir: pathlib.Path) -> pathlib.Path:
    return test_dir / "functional" / "fixtures"


@pytest.fixture(scope="session")
def gitlab_service_name() -> str:
    """The "service" name is the one defined in the `docker-compose.yml` file"""
    return "gitlab"


@pytest.fixture(scope="session")
def gitlab_container_name() -> str:
    """The "container" name is the one defined in the `docker-compose.yml` file
    for the "gitlab" service"""
    return "gitlab-test"


@pytest.fixture(scope="session")
def gitlab_docker_port(docker_services, gitlab_service_name: str) -> int:
    port: int = docker_services.port_for(gitlab_service_name, container_port=80)
    return port


@pytest.fixture(scope="session")
def gitlab_url(docker_ip: str, gitlab_docker_port: int) -> str:
    return f"http://{docker_ip}:{gitlab_docker_port}"


def reset_gitlab(gl: gitlab.Gitlab) -> None:
    """Delete resources (such as projects, groups, users) that shouldn't
    exist."""
    if helpers.get_gitlab_plan(gl):
        logging.info("GitLab EE detected")
        # NOTE(jlvillal, timknight): By default in GitLab EE it will wait 7 days before
        # deleting a group or project.
        # In GL 16.0 we need to call delete with `permanently_remove=True` for projects and sub groups
        # (handled in helpers.py safe_delete)
        settings = gl.settings.get()
        modified_settings = False
        if settings.deletion_adjourned_period != 1:
            logging.info("Setting `deletion_adjourned_period` to 1 Day")
            settings.deletion_adjourned_period = 1
            modified_settings = True
        if modified_settings:
            settings.save()

    for project in gl.projects.list():
        for deploy_token in project.deploytokens.list():
            logging.info(
                f"Deleting deploy token: {deploy_token.username!r} in "
                f"project: {project.path_with_namespace!r}"
            )
            helpers.safe_delete(deploy_token)
        logging.info(f"Deleting project: {project.path_with_namespace!r}")
        helpers.safe_delete(project)

    for group in gl.groups.list():
        # skip deletion of a descendant group to prevent scenarios where parent group
        # gets deleted leaving a dangling descendant whose deletion will throw 404s.
        if group.parent_id:
            logging.info(
                f"Skipping deletion of {group.full_path} as it is a descendant "
                f"group and will be removed when the parent group is deleted"
            )
            continue

        for deploy_token in group.deploytokens.list():
            logging.info(
                f"Deleting deploy token: {deploy_token.username!r} in "
                f"group: {group.path_with_namespace!r}"
            )
            helpers.safe_delete(deploy_token)
        logging.info(f"Deleting group: {group.full_path!r}")
        helpers.safe_delete(group)
    for topic in gl.topics.list():
        logging.info(f"Deleting topic: {topic.name!r}")
        helpers.safe_delete(topic)
    for variable in gl.variables.list():
        logging.info(f"Deleting variable: {variable.key!r}")
        helpers.safe_delete(variable)
    for user in gl.users.list():
        if user.username not in ["root", "ghost"]:
            logging.info(f"Deleting user: {user.username!r}")
            helpers.safe_delete(user)


def set_token(container: str, fixture_dir: pathlib.Path) -> str:
    logging.info("Creating API token.")
    set_token_rb = fixture_dir / "set_token.rb"

    with open(set_token_rb, "r", encoding="utf-8") as f:
        set_token_command = f.read().strip()

    rails_command = [
        "docker",
        "exec",
        container,
        "gitlab-rails",
        "runner",
        set_token_command,
    ]
    output = check_output(rails_command).decode().strip()
    logging.info("Finished creating API token.")

    return output


def pytest_report_collectionfinish(config, startdir, items):
    return [
        "",
        "Starting GitLab container.",
        "Waiting for GitLab to reconfigure.",
        "This will take a few minutes.",
    ]


def pytest_addoption(parser):
    parser.addoption(
        "--keep-containers",
        action="store_true",
        help="Keep containers running after testing",
    )


@pytest.fixture(scope="session")
def temp_dir() -> pathlib.Path:
    return pathlib.Path(tempfile.gettempdir())


@pytest.fixture(scope="session")
def check_is_alive():
    """
    Return a healthcheck function fixture for the GitLab container spinup.
    """

    def _check(
        *,
        container: str,
        start_time: float,
        gitlab_url: str,
    ) -> bool:
        setup_time = time.perf_counter() - start_time
        minutes, seconds = int(setup_time / 60), int(setup_time % 60)
        logging.info(
            f"Checking if GitLab container is up. "
            f"Have been checking for {minutes} minute(s), {seconds} seconds ..."
        )
        logs = ["docker", "logs", container]
        if "gitlab Reconfigured!" not in check_output(logs).decode():
            return False
        logging.debug("GitLab has finished reconfiguring.")
        for check in ("health", "readiness", "liveness"):
            url = f"{gitlab_url}/-/{check}"
            logging.debug(f"Checking {check!r} endpoint at: {url}")
            try:
                result = requests.get(url, timeout=1.0)
            except requests.exceptions.Timeout:
                logging.info(f"{check!r} check timed out")
                return False
            if result.status_code != 200:
                logging.info(f"{check!r} check did not return 200: {result!r}")
                return False
            logging.debug(f"{check!r} check passed: {result!r}")
        logging.debug(f"Sleeping for {SLEEP_TIME}")
        time.sleep(SLEEP_TIME)
        return True

    return _check


@pytest.fixture(scope="session")
def gitlab_token(
    check_is_alive,
    gitlab_container_name: str,
    gitlab_url: str,
    docker_services,
    fixture_dir: pathlib.Path,
) -> str:
    start_time = time.perf_counter()
    logging.info("Waiting for GitLab container to become ready.")
    docker_services.wait_until_responsive(
        timeout=300,
        pause=10,
        check=lambda: check_is_alive(
            container=gitlab_container_name,
            start_time=start_time,
            gitlab_url=gitlab_url,
        ),
    )
    setup_time = time.perf_counter() - start_time
    minutes, seconds = int(setup_time / 60), int(setup_time % 60)
    logging.info(
        f"GitLab container is now ready after {minutes} minute(s), {seconds} seconds"
    )

    return set_token(gitlab_container_name, fixture_dir=fixture_dir)


@pytest.fixture(scope="session")
def gitlab_config(gitlab_url: str, gitlab_token: str, temp_dir: pathlib.Path):
    config_file = temp_dir / "python-gitlab.cfg"

    config = f"""[global]
default = local
timeout = 60

[local]
url = {gitlab_url}
private_token = {gitlab_token}
api_version = 4"""

    with open(config_file, "w", encoding="utf-8") as f:
        f.write(config)

    return config_file


@pytest.fixture(scope="session")
def gl(gitlab_url: str, gitlab_token: str) -> gitlab.Gitlab:
    """Helper instance to make fixtures and asserts directly via the API."""

    logging.info("Instantiating python-gitlab gitlab.Gitlab instance")
    instance = gitlab.Gitlab(gitlab_url, private_token=gitlab_token)

    logging.info("Reset GitLab")
    reset_gitlab(instance)

    return instance


@pytest.fixture(scope="session")
def gitlab_plan(gl: gitlab.Gitlab) -> Optional[str]:
    return helpers.get_gitlab_plan(gl)


@pytest.fixture(autouse=True)
def gitlab_premium(gitlab_plan, request) -> None:
    if gitlab_plan in ("premium", "ultimate"):
        return

    if request.node.get_closest_marker("gitlab_ultimate"):
        pytest.skip("Test requires GitLab Premium plan")


@pytest.fixture(autouse=True)
def gitlab_ultimate(gitlab_plan, request) -> None:
    if gitlab_plan == "ultimate":
        return

    if request.node.get_closest_marker("gitlab_ultimate"):
        pytest.skip("Test requires GitLab Ultimate plan")


@pytest.fixture(scope="session")
def gitlab_runner(gl):
    container = "gitlab-runner-test"
    runner_name = "python-gitlab-runner"
    token = "registration-token"
    url = "http://gitlab"

    docker_exec = ["docker", "exec", container, "gitlab-runner"]
    register = [
        "register",
        "--run-untagged",
        "--non-interactive",
        "--registration-token",
        token,
        "--name",
        runner_name,
        "--url",
        url,
        "--clone-url",
        url,
        "--executor",
        "shell",
    ]
    unregister = ["unregister", "--name", runner_name]

    yield check_output(docker_exec + register).decode()

    check_output(docker_exec + unregister).decode()


@pytest.fixture(scope="module")
def group(gl):
    """Group fixture for group API resource tests."""
    _id = uuid.uuid4().hex
    data = {
        "name": f"test-group-{_id}",
        "path": f"group-{_id}",
    }
    group = gl.groups.create(data)

    yield group

    helpers.safe_delete(group)


@pytest.fixture(scope="module")
def project(gl):
    """Project fixture for project API resource tests."""
    _id = uuid.uuid4().hex
    name = f"test-project-{_id}"

    project = gl.projects.create(name=name)

    yield project

    helpers.safe_delete(project)


@pytest.fixture(scope="function")
def make_merge_request(project):
    """Fixture factory used to create a merge_request.

    It will create a branch, add a commit to the branch, and then create a
    merge request against project.default_branch. The MR will be returned.

    When finished any created merge requests and branches will be deleted.

    NOTE: No attempt is made to restore project.default_branch to its previous
    state. So if the merge request is merged then its content will be in the
    project.default_branch branch.
    """

    to_delete = []

    def _make_merge_request(*, source_branch: str, create_pipeline: bool = False):
        # Wait for processes to be done before we start...
        # NOTE(jlvillal): Sometimes the CI would give a "500 Internal Server
        # Error". Hoping that waiting until all other processes are done will
        # help with that.
        # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
        time.sleep(30)

        project.refresh()  # Gets us the current default branch
        logging.info(f"Creating branch {source_branch}")
        mr_branch = project.branches.create(
            {"branch": source_branch, "ref": project.default_branch}
        )
        # NOTE(jlvillal): Must create a commit in the new branch before we can
        # create an MR that will work.
        project.files.create(
            {
                "file_path": f"README.{source_branch}",
                "branch": source_branch,
                "content": "Initial content",
                "commit_message": "New commit in new branch",
            }
        )

        if create_pipeline:
            project.files.create(
                {
                    "file_path": ".gitlab-ci.yml",
                    "branch": source_branch,
                    "content": """
test:
  rules:
    - if: '$CI_PIPELINE_SOURCE == "merge_request_event"'
  script:
    - sleep 24h  # We don't expect this to finish
""",
                    "commit_message": "Add a simple pipeline",
                }
            )
        mr = project.mergerequests.create(
            {
                "source_branch": source_branch,
                "target_branch": project.default_branch,
                "title": "Should remove source branch",
                "remove_source_branch": True,
            }
        )

        # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
        time.sleep(5)

        mr_iid = mr.iid
        for _ in range(60):
            mr = project.mergerequests.get(mr_iid)
            if (
                mr.detailed_merge_status == "checking"
                or mr.detailed_merge_status == "unchecked"
            ):
                time.sleep(0.5)
            else:
                break

        assert mr.detailed_merge_status != "checking"
        assert mr.detailed_merge_status != "unchecked"

        to_delete.extend([mr, mr_branch])
        return mr

    yield _make_merge_request

    for object in to_delete:
        helpers.safe_delete(object)


@pytest.fixture(scope="function")
def merge_request(make_merge_request, project):
    _id = uuid.uuid4().hex
    return make_merge_request(source_branch=f"branch-{_id}")


@pytest.fixture(scope="function")
def merge_request_with_pipeline(make_merge_request, project):
    _id = uuid.uuid4().hex
    return make_merge_request(source_branch=f"branch-{_id}", create_pipeline=True)


@pytest.fixture(scope="module")
def project_file(project):
    """File fixture for tests requiring a project with files and branches."""
    project_file = project.files.create(
        {
            "file_path": "README",
            "branch": "main",
            "content": "Initial content",
            "commit_message": "Initial commit",
        }
    )

    return project_file


@pytest.fixture(scope="function")
def release(project, project_file):
    _id = uuid.uuid4().hex
    name = f"we_have_a_slash/test-release-{_id}"

    project.refresh()  # Gets us the current default branch
    release = project.releases.create(
        {
            "name": name,
            "tag_name": _id,
            "description": "description",
            "ref": project.default_branch,
        }
    )

    return release


@pytest.fixture(scope="function")
def service(project):
    """This is just a convenience fixture to make test cases slightly prettier. Project
    services are not idempotent. A service cannot be retrieved until it is enabled.
    After it is enabled the first time, it can never be fully deleted, only disabled."""
    service = project.services.update("asana", {"api_key": "api_key"})

    yield service

    try:
        project.services.delete("asana")
    except gitlab.exceptions.GitlabDeleteError as e:
        print(f"Service already disabled: {e}")


@pytest.fixture(scope="module")
def user(gl):
    """User fixture for user API resource tests."""
    _id = uuid.uuid4().hex
    email = f"user{_id}@email.com"
    username = f"user{_id}"
    name = f"User {_id}"
    password = "E4596f8be406Bc3a14a4ccdb1df80587"

    user = gl.users.create(email=email, username=username, name=name, password=password)

    yield user

    helpers.safe_delete(user)


@pytest.fixture(scope="module")
def issue(project):
    """Issue fixture for issue API resource tests."""
    _id = uuid.uuid4().hex
    data = {"title": f"Issue {_id}", "description": f"Issue {_id} description"}

    return project.issues.create(data)


@pytest.fixture(scope="module")
def milestone(project):
    _id = uuid.uuid4().hex
    data = {"title": f"milestone{_id}"}

    return project.milestones.create(data)


@pytest.fixture(scope="module")
def label(project):
    """Label fixture for project label API resource tests."""
    _id = uuid.uuid4().hex
    data = {
        "name": f"prjlabel{_id}",
        "description": f"prjlabel1 {_id} description",
        "color": "#112233",
    }

    return project.labels.create(data)


@pytest.fixture(scope="module")
def group_label(group):
    """Label fixture for group label API resource tests."""
    _id = uuid.uuid4().hex
    data = {
        "name": f"grplabel{_id}",
        "description": f"grplabel1 {_id} description",
        "color": "#112233",
    }

    return group.labels.create(data)


@pytest.fixture(scope="module")
def epic(group):
    """Fixture for group epic API resource tests."""
    _id = uuid.uuid4().hex
    return group.epics.create({"title": f"epic-{_id}", "description": f"Epic {_id}"})


@pytest.fixture(scope="module")
def variable(project):
    """Variable fixture for project variable API resource tests."""
    _id = uuid.uuid4().hex
    data = {"key": f"var{_id}", "value": f"Variable {_id}"}

    return project.variables.create(data)


@pytest.fixture(scope="module")
def deploy_token(project):
    """Deploy token fixture for project deploy token API resource tests."""
    _id = uuid.uuid4().hex
    data = {
        "name": f"token-{_id}",
        "username": "root",
        "expires_at": datetime.date.today().isoformat(),
        "scopes": "read_registry",
    }

    return project.deploytokens.create(data)


@pytest.fixture(scope="module")
def group_deploy_token(group):
    """Deploy token fixture for group deploy token API resource tests."""
    _id = uuid.uuid4().hex
    data = {
        "name": f"group-token-{_id}",
        "username": "root",
        "expires_at": datetime.date.today().isoformat(),
        "scopes": "read_registry",
    }

    return group.deploytokens.create(data)


@pytest.fixture(scope="session")
def GPG_KEY():
    return """-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBFn5mzYBCADH6SDVPAp1zh/hxmTi0QplkOfExBACpuY6OhzNdIg+8/528b3g
Y5YFR6T/HLv/PmeHskUj21end1C0PNG2T9dTx+2Vlh9ISsSG1kyF9T5fvMR3bE0x
Dl6S489CXZrjPTS9SHk1kF+7dwjUxLJyxF9hPiSihFefDFu3NeOtG/u8vbC1mewQ
ZyAYue+mqtqcCIFFoBz7wHKMWjIVSJSyTkXExu4OzpVvy3l2EikbvavI3qNz84b+
Mgkv/kiBlNoCy3CVuPk99RYKZ3lX1vVtqQ0OgNGQvb4DjcpyjmbKyibuZwhDjIOh
au6d1OyEbayTntd+dQ4j9EMSnEvm/0MJ4eXPABEBAAG0G0dpdGxhYlRlc3QxIDxm
YWtlQGZha2UudGxkPokBNwQTAQgAIQUCWfmbNgIbAwULCQgHAgYVCAkKCwIEFgID
AQIeAQIXgAAKCRBgxELHf8f3hF3yB/wNJlWPKY65UsB4Lo0hs1OxdxCDqXogSi0u
6crDEIiyOte62pNZKzWy8TJcGZvznRTZ7t8hXgKFLz3PRMcl+vAiRC6quIDUj+2V
eYfwaItd1lUfzvdCaC7Venf4TQ74f5vvNg/zoGwE6eRoSbjlLv9nqsxeA0rUBUQL
LYikWhVMP3TrlfgfduYvh6mfgh57BDLJ9kJVpyfxxx9YLKZbaas9sPa6LgBtR555
JziUxHmbEv8XCsUU8uoFeP1pImbNBplqE3wzJwzOMSmmch7iZzrAwfN7N2j3Wj0H
B5kQddJ9dmB4BbU0IXGhWczvdpxboI2wdY8a1JypxOdePoph/43iuQENBFn5mzYB
CADnTPY0Zf3d9zLjBNgIb3yDl94uOcKCq0twNmyjMhHzGqw+UMe9BScy34GL94Al
xFRQoaL+7P8hGsnsNku29A/VDZivcI+uxTx4WQ7OLcn7V0bnHV4d76iky2ufbUt/
GofthjDs1SonePO2N09sS4V4uK0d5N4BfCzzXgvg8etCLxNmC9BGt7AaKUUzKBO4
2QvNNaC2C/8XEnOgNWYvR36ylAXAmo0sGFXUsBCTiq1fugS9pwtaS2JmaVpZZ3YT
pMZlS0+SjC5BZYFqSmKCsA58oBRzCxQz57nR4h5VEflgD+Hy0HdW0UHETwz83E6/
U0LL6YyvhwFr6KPq5GxinSvfABEBAAGJAR8EGAEIAAkFAln5mzYCGwwACgkQYMRC
x3/H94SJgwgAlKQb10/xcL/epdDkR7vbiei7huGLBpRDb/L5fM8B5W77Qi8Xmuqj
cCu1j99ZCA5hs/vwVn8j8iLSBGMC5gxcuaar/wtmiaEvT9fO/h6q4opG7NcuiJ8H
wRj8ccJmRssNqDD913PLz7T40Ts62blhrEAlJozGVG/q7T3RAZcskOUHKeHfc2RI
YzGsC/I9d7k6uxAv1L9Nm5F2HaAQDzhkdd16nKkGaPGR35cT1JLInkfl5cdm7ldN
nxs4TLO3kZjUTgWKdhpgRNF5hwaz51ZjpebaRf/ZqRuNyX4lIRolDxzOn/+O1o8L
qG2ZdhHHmSK2LaQLFiSprUkikStNU9BqSQ==
=5OGa
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture(scope="session")
def SSH_KEY():
    return (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDZAjAX8vTiHD7Yi3/EzuVaDChtih"
        "79HyJZ6H9dEqxFfmGA1YnncE0xujQ64TCebhkYJKzmTJCImSVkOu9C4hZgsw6eE76n"
        "+Cg3VwEeDUFy+GXlEJWlHaEyc3HWioxgOALbUp3rOezNh+d8BDwwqvENGoePEBsz5l"
        "a6WP5lTi/HJIjAl6Hu+zHgdj1XVExeH+S52EwpZf/ylTJub0Bl5gHwf/siVE48mLMI"
        "sqrukXTZ6Zg+8EHAIvIQwJ1dKcXe8P5IoLT7VKrbkgAnolS0I8J+uH7KtErZJb5oZh"
        "S4OEwsNpaXMAr+6/wWSpircV2/e7sFLlhlKBC4Iq1MpqlZ7G3p foo@bar"
    )


@pytest.fixture(scope="session")
def DEPLOY_KEY():
    return (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDFdRyjJQh+1niBpXqE2I8dzjG"
        "MXFHlRjX9yk/UfOn075IdaockdU58sw2Ai1XIWFpZpfJkW7z+P47ZNSqm1gzeXI"
        "rtKa9ZUp8A7SZe8vH4XVn7kh7bwWCUirqtn8El9XdqfkzOs/+FuViriUWoJVpA6"
        "WZsDNaqINFKIA5fj/q8XQw+BcS92L09QJg9oVUuH0VVwNYbU2M2IRmSpybgC/gu"
        "uWTrnCDMmLItksATifLvRZwgdI8dr+q6tbxbZknNcgEPrI2jT0hYN9ZcjNeWuyv"
        "rke9IepE7SPBT41C+YtUX4dfDZDmczM1cE0YL/krdUCfuZHMa4ZS2YyNd6slufc"
        "vn bar@foo"
    )



"""
GitLab API: https://docs.gitlab.com/ee/api/iterations.html
"""

import re

import pytest
import responses

iterations_content = [
    {
        "id": 53,
        "iid": 13,
        "group_id": 5,
        "title": "Iteration II",
        "description": "Ipsum Lorem ipsum",
        "state": 2,
        "created_at": "2020-01-27T05:07:12.573Z",
        "updated_at": "2020-01-27T05:07:12.573Z",
        "due_date": "2020-02-01",
        "start_date": "2020-02-14",
        "web_url": "http://gitlab.example.com/groups/my-group/-/iterations/13",
    }
]


@pytest.fixture
def resp_iterations_list():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/iterations"),
            json=iterations_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_group_iterations(group, resp_iterations_list):
    iterations = group.iterations.list()
    assert iterations[0].group_id == 5


def test_list_project_iterations(project, resp_iterations_list):
    iterations = project.iterations.list()
    assert iterations[0].group_id == 5



"""
Gitlab API: https://docs.gitlab.com/ee/api/merge_request_approvals.html
"""

import copy

import pytest
import responses

import gitlab
from gitlab.mixins import UpdateMethod

approval_rule_id = 7
approval_rule_name = "security"
approvals_required = 3
user_ids = [5, 50]
group_ids = [5]

new_approval_rule_name = "new approval rule"
new_approval_rule_user_ids = user_ids
new_approval_rule_approvals_required = 2

updated_approval_rule_user_ids = [5]
updated_approval_rule_approvals_required = 1


@pytest.fixture
def resp_mr_approval_rules():
    mr_ars_content = [
        {
            "id": approval_rule_id,
            "name": approval_rule_name,
            "rule_type": "regular",
            "eligible_approvers": [
                {
                    "id": user_ids[0],
                    "name": "John Doe",
                    "username": "jdoe",
                    "state": "active",
                    "avatar_url": "https://www.gravatar.com/avatar/0?s=80&d=identicon",
                    "web_url": "http://localhost/jdoe",
                },
                {
                    "id": user_ids[1],
                    "name": "Group Member 1",
                    "username": "group_member_1",
                    "state": "active",
                    "avatar_url": "https://www.gravatar.com/avatar/0?s=80&d=identicon",
                    "web_url": "http://localhost/group_member_1",
                },
            ],
            "approvals_required": approvals_required,
            "source_rule": None,
            "users": [
                {
                    "id": 5,
                    "name": "John Doe",
                    "username": "jdoe",
                    "state": "active",
                    "avatar_url": "https://www.gravatar.com/avatar/0?s=80&d=identicon",
                    "web_url": "http://localhost/jdoe",
                }
            ],
            "groups": [
                {
                    "id": 5,
                    "name": "group1",
                    "path": "group1",
                    "description": "",
                    "visibility": "public",
                    "lfs_enabled": False,
                    "avatar_url": None,
                    "web_url": "http://localhost/groups/group1",
                    "request_access_enabled": False,
                    "full_name": "group1",
                    "full_path": "group1",
                    "parent_id": None,
                    "ldap_cn": None,
                    "ldap_access": None,
                }
            ],
            "contains_hidden_groups": False,
            "overridden": False,
        }
    ]

    approval_state_rules = copy.deepcopy(mr_ars_content)
    approval_state_rules[0]["approved"] = False
    approval_state_rules[0]["approved_by"] = []

    mr_approval_state_content = {
        "approval_rules_overwritten": False,
        "rules": approval_state_rules,
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_rules",
            json=mr_ars_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_rules/7",
            json=mr_ars_content[0],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_state",
            json=mr_approval_state_content,
            content_type="application/json",
            status=200,
        )

        new_mr_ars_content = dict(mr_ars_content[0])
        new_mr_ars_content["name"] = new_approval_rule_name
        new_mr_ars_content["approvals_required"] = new_approval_rule_approvals_required

        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_rules",
            json=new_mr_ars_content,
            content_type="application/json",
            status=200,
        )

        updated_mr_ars_content = copy.deepcopy(mr_ars_content[0])
        updated_mr_ars_content["eligible_approvers"] = [
            mr_ars_content[0]["eligible_approvers"][0]
        ]

        updated_mr_ars_content["approvals_required"] = (
            updated_approval_rule_approvals_required
        )

        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_rules/7",
            json=updated_mr_ars_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_mr_approval_rule():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/merge_requests/3/approval_rules/7",
            status=204,
        )
        yield rsps


def test_project_approval_manager_update_method_post(project):
    """Ensure the
    gitlab.v4.objects.merge_request_approvals.ProjectApprovalManager object has
    _update_method set to UpdateMethod.POST"""
    approvals = project.approvals
    assert isinstance(
        approvals, gitlab.v4.objects.merge_request_approvals.ProjectApprovalManager
    )
    assert approvals._update_method is UpdateMethod.POST


def test_list_merge_request_approval_rules(project, resp_mr_approval_rules):
    approval_rules = project.mergerequests.get(3, lazy=True).approval_rules.list()
    assert len(approval_rules) == 1
    assert approval_rules[0].name == approval_rule_name
    assert approval_rules[0].id == approval_rule_id
    repr(approval_rules)  # ensure that `repr()` doesn't raise an exception


def test_delete_merge_request_approval_rule(project, resp_delete_mr_approval_rule):
    merge_request = project.mergerequests.get(3, lazy=True)
    merge_request.approval_rules.delete(approval_rule_id)


def test_update_merge_request_approvals_set_approvers(project, resp_mr_approval_rules):
    approvals = project.mergerequests.get(3, lazy=True).approvals
    assert isinstance(
        approvals,
        gitlab.v4.objects.merge_request_approvals.ProjectMergeRequestApprovalManager,
    )
    assert approvals._update_method is UpdateMethod.POST
    response = approvals.set_approvers(
        updated_approval_rule_approvals_required,
        approver_ids=updated_approval_rule_user_ids,
        approver_group_ids=group_ids,
        approval_rule_name=approval_rule_name,
    )

    assert response.approvals_required == updated_approval_rule_approvals_required
    assert len(response.eligible_approvers) == len(updated_approval_rule_user_ids)
    assert response.eligible_approvers[0]["id"] == updated_approval_rule_user_ids[0]
    assert response.name == approval_rule_name


def test_create_merge_request_approvals_set_approvers(project, resp_mr_approval_rules):
    approvals = project.mergerequests.get(3, lazy=True).approvals
    assert isinstance(
        approvals,
        gitlab.v4.objects.merge_request_approvals.ProjectMergeRequestApprovalManager,
    )
    assert approvals._update_method is UpdateMethod.POST
    response = approvals.set_approvers(
        new_approval_rule_approvals_required,
        approver_ids=new_approval_rule_user_ids,
        approver_group_ids=group_ids,
        approval_rule_name=new_approval_rule_name,
    )
    assert response.approvals_required == new_approval_rule_approvals_required
    assert len(response.eligible_approvers) == len(new_approval_rule_user_ids)
    assert response.eligible_approvers[0]["id"] == new_approval_rule_user_ids[0]
    assert response.name == new_approval_rule_name


def test_create_merge_request_approval_rule(project, resp_mr_approval_rules):
    approval_rules = project.mergerequests.get(3, lazy=True).approval_rules
    data = {
        "name": new_approval_rule_name,
        "approvals_required": new_approval_rule_approvals_required,
        "rule_type": "regular",
        "user_ids": new_approval_rule_user_ids,
        "group_ids": group_ids,
    }
    response = approval_rules.create(data)
    assert response.approvals_required == new_approval_rule_approvals_required
    assert len(response.eligible_approvers) == len(new_approval_rule_user_ids)
    assert response.eligible_approvers[0]["id"] == new_approval_rule_user_ids[0]
    assert response.name == new_approval_rule_name


def test_update_merge_request_approval_rule(project, resp_mr_approval_rules):
    approval_rules = project.mergerequests.get(3, lazy=True).approval_rules
    ar_1 = approval_rules.list()[0]
    ar_1.user_ids = updated_approval_rule_user_ids
    ar_1.approvals_required = updated_approval_rule_approvals_required
    ar_1.save()

    assert ar_1.approvals_required == updated_approval_rule_approvals_required
    assert len(ar_1.eligible_approvers) == len(updated_approval_rule_user_ids)
    assert ar_1.eligible_approvers[0]["id"] == updated_approval_rule_user_ids[0]


def test_get_merge_request_approval_rule(project, resp_mr_approval_rules):
    merge_request = project.mergerequests.get(3, lazy=True)
    approval_rule = merge_request.approval_rules.get(approval_rule_id)
    assert isinstance(
        approval_rule,
        gitlab.v4.objects.merge_request_approvals.ProjectMergeRequestApprovalRule,
    )
    assert approval_rule.name == approval_rule_name
    assert approval_rule.id == approval_rule_id


def test_get_merge_request_approval_state(project, resp_mr_approval_rules):
    merge_request = project.mergerequests.get(3, lazy=True)
    approval_state = merge_request.approval_state.get()
    assert isinstance(
        approval_state,
        gitlab.v4.objects.merge_request_approvals.ProjectMergeRequestApprovalState,
    )
    assert not approval_state.approval_rules_overwritten
    assert len(approval_state.rules) == 1
    assert approval_state.rules[0]["name"] == approval_rule_name
    assert approval_state.rules[0]["id"] == approval_rule_id
    assert not approval_state.rules[0]["approved"]
    assert approval_state.rules[0]["approved_by"] == []



"""
GitLab API: https://docs.gitlab.com/ee/api/group_access_tokens.html
"""

import pytest
import responses

from gitlab.v4.objects import GroupAccessToken


@pytest.fixture
def resp_list_group_access_token(token_content):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/access_tokens",
            json=[token_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_group_access_token(token_content):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/access_tokens/1",
            json=token_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_group_access_token(token_content):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/access_tokens",
            json=token_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_revoke_group_access_token():
    content = [
        {
            "user_id": 141,
            "scopes": ["api"],
            "name": "token",
            "expires_at": "2021-01-31",
            "id": 42,
            "active": True,
            "created_at": "2021-01-20T22:11:48.151Z",
            "revoked": False,
        }
    ]

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/groups/1/access_tokens/42",
            status=204,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/access_tokens",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_rotate_group_access_token(token_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/access_tokens/1/rotate",
            json=token_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_group_access_tokens(gl, resp_list_group_access_token):
    access_tokens = gl.groups.get(1, lazy=True).access_tokens.list()
    assert len(access_tokens) == 1
    assert access_tokens[0].revoked is False
    assert access_tokens[0].name == "token"


def test_get_group_access_token(group, resp_get_group_access_token):
    access_token = group.access_tokens.get(1)
    assert isinstance(access_token, GroupAccessToken)
    assert access_token.revoked is False
    assert access_token.name == "token"


def test_create_group_access_token(gl, resp_create_group_access_token):
    access_tokens = gl.groups.get(1, lazy=True).access_tokens.create(
        {"name": "test", "scopes": ["api"]}
    )
    assert access_tokens.revoked is False
    assert access_tokens.user_id == 141
    assert access_tokens.expires_at == "2021-01-31"


def test_revoke_group_access_token(
    gl, resp_list_group_access_token, resp_revoke_group_access_token
):
    gl.groups.get(1, lazy=True).access_tokens.delete(42)
    access_token = gl.groups.get(1, lazy=True).access_tokens.list()[0]
    access_token.delete()


def test_rotate_group_access_token(group, resp_rotate_group_access_token):
    access_token = group.access_tokens.get(1, lazy=True)
    access_token.rotate()
    assert isinstance(access_token, GroupAccessToken)
    assert access_token.token == "s3cr3t"



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import DeleteMixin, ObjectDeleteMixin, RetrieveMixin, SetMixin

__all__ = [
    "GroupCustomAttribute",
    "GroupCustomAttributeManager",
    "ProjectCustomAttribute",
    "ProjectCustomAttributeManager",
    "UserCustomAttribute",
    "UserCustomAttributeManager",
]


class GroupCustomAttribute(ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class GroupCustomAttributeManager(RetrieveMixin, SetMixin, DeleteMixin, RESTManager):
    _path = "/groups/{group_id}/custom_attributes"
    _obj_cls = GroupCustomAttribute
    _from_parent_attrs = {"group_id": "id"}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> GroupCustomAttribute:
        return cast(GroupCustomAttribute, super().get(id=id, lazy=lazy, **kwargs))


class ProjectCustomAttribute(ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class ProjectCustomAttributeManager(RetrieveMixin, SetMixin, DeleteMixin, RESTManager):
    _path = "/projects/{project_id}/custom_attributes"
    _obj_cls = ProjectCustomAttribute
    _from_parent_attrs = {"project_id": "id"}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectCustomAttribute:
        return cast(ProjectCustomAttribute, super().get(id=id, lazy=lazy, **kwargs))


class UserCustomAttribute(ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class UserCustomAttributeManager(RetrieveMixin, SetMixin, DeleteMixin, RESTManager):
    _path = "/users/{user_id}/custom_attributes"
    _obj_cls = UserCustomAttribute
    _from_parent_attrs = {"user_id": "id"}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> UserCustomAttribute:
        return cast(UserCustomAttribute, super().get(id=id, lazy=lazy, **kwargs))



import datetime


def test_create_user_impersonation_token_with_scopes(gitlab_cli, user):
    cmd = [
        "user-impersonation-token",
        "create",
        "--user-id",
        user.id,
        "--name",
        "test-token",
        "--scopes",
        "api,read_user",
        "--expires-at",
        datetime.date.today().isoformat(),
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_user_projects(gitlab_cli, user):
    cmd = ["user-project", "list", "--user-id", user.id]
    ret = gitlab_cli(cmd)

    assert ret.success



"""
GitLab API: https://docs.gitlab.com/ee/api/draft_notes.html
"""

from copy import deepcopy

import pytest
import responses

from gitlab.v4.objects import ProjectMergeRequestDraftNote

draft_note_content = {
    "id": 1,
    "author_id": 23,
    "merge_request_id": 1,
    "resolve_discussion": False,
    "discussion_id": None,
    "note": "Example title",
    "commit_id": None,
    "line_code": None,
    "position": {
        "base_sha": None,
        "start_sha": None,
        "head_sha": None,
        "old_path": None,
        "new_path": None,
        "position_type": "text",
        "old_line": None,
        "new_line": None,
        "line_range": None,
    },
}


@pytest.fixture()
def resp_list_merge_request_draft_notes():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes",
            json=[draft_note_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_get_merge_request_draft_note():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes/1",
            json=draft_note_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_create_merge_request_draft_note():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes",
            json=draft_note_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture()
def resp_update_merge_request_draft_note():
    updated_content = deepcopy(draft_note_content)
    updated_content["note"] = "New title"

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes/1",
            json=updated_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture()
def resp_delete_merge_request_draft_note():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes/1",
            json=draft_note_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture()
def resp_publish_merge_request_draft_note():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes/1/publish",
            status=204,
        )
        yield rsps


@pytest.fixture()
def resp_bulk_publish_merge_request_draft_notes():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/merge_requests/1/draft_notes/bulk_publish",
            status=204,
        )
        yield rsps


def test_list_merge_requests_draft_notes(
    project_merge_request, resp_list_merge_request_draft_notes
):
    draft_notes = project_merge_request.draft_notes.list()
    assert len(draft_notes) == 1
    assert isinstance(draft_notes[0], ProjectMergeRequestDraftNote)
    assert draft_notes[0].note == draft_note_content["note"]


def test_get_merge_requests_draft_note(
    project_merge_request, resp_get_merge_request_draft_note
):
    draft_note = project_merge_request.draft_notes.get(1)
    assert isinstance(draft_note, ProjectMergeRequestDraftNote)
    assert draft_note.note == draft_note_content["note"]


def test_create_merge_requests_draft_note(
    project_merge_request, resp_create_merge_request_draft_note
):
    draft_note = project_merge_request.draft_notes.create({"note": "Example title"})
    assert isinstance(draft_note, ProjectMergeRequestDraftNote)
    assert draft_note.note == draft_note_content["note"]


def test_update_merge_requests_draft_note(
    project_merge_request, resp_update_merge_request_draft_note
):
    draft_note = project_merge_request.draft_notes.get(1, lazy=True)
    draft_note.note = "New title"
    draft_note.save()
    assert draft_note.note == "New title"


def test_delete_merge_requests_draft_note(
    project_merge_request, resp_delete_merge_request_draft_note
):
    draft_note = project_merge_request.draft_notes.get(1, lazy=True)
    draft_note.delete()


def test_publish_merge_requests_draft_note(
    project_merge_request, resp_publish_merge_request_draft_note
):
    draft_note = project_merge_request.draft_notes.get(1, lazy=True)
    draft_note.publish()


def test_bulk_publish_merge_requests_draft_notes(
    project_merge_request, resp_bulk_publish_merge_request_draft_notes
):
    project_merge_request.draft_notes.bulk_publish()



import time
import uuid

import pytest

import gitlab
from gitlab.const import AccessLevel
from gitlab.v4.objects.projects import ProjectStorage


def test_projects_head(gl):
    headers = gl.projects.head()
    assert headers["x-total"]


def test_project_head(gl, project):
    headers = gl.projects.head(project.id)
    assert headers["content-type"] == "application/json"


def test_create_project(gl, user):
    # Moved from group tests chunk in legacy tests, TODO cleanup
    admin_project = gl.projects.create({"name": "admin_project"})
    assert isinstance(admin_project, gitlab.v4.objects.Project)
    assert admin_project in gl.projects.list(search="admin_project")

    sudo_project = gl.projects.create({"name": "sudo_project"}, sudo=user.id)

    created = gl.projects.list()
    created_gen = gl.projects.list(iterator=True)
    owned = gl.projects.list(owned=True)

    assert admin_project in created and sudo_project in created
    assert admin_project in owned and sudo_project not in owned
    assert len(created) == len(list(created_gen))

    admin_project.delete()
    sudo_project.delete()


def test_project_members(user, project):
    member = project.members.create(
        {"user_id": user.id, "access_level": AccessLevel.DEVELOPER}
    )
    assert member in project.members.list()
    assert member.access_level == 30

    member.delete()


def test_project_badges(project):
    badge_image = "http://example.com"
    badge_link = "http://example/img.svg"

    badge = project.badges.create({"link_url": badge_link, "image_url": badge_image})
    assert badge in project.badges.list()

    badge.image_url = "http://another.example.com"
    badge.save()

    badge = project.badges.get(badge.id)
    assert badge.image_url == "http://another.example.com"

    badge.delete()


@pytest.mark.skip(reason="Commented out in legacy test")
def test_project_boards(project):
    boards = project.boards.list()
    assert boards

    board = boards[0]
    lists = board.lists.list()

    last_list = lists[-1]
    last_list.position = 0
    last_list.save()

    last_list.delete()


def test_project_custom_attributes(gl, project):
    attrs = project.customattributes.list()
    assert not attrs

    attr = project.customattributes.set("key", "value1")
    assert attr.key == "key"
    assert attr.value == "value1"
    assert attr in project.customattributes.list()
    assert project in gl.projects.list(custom_attributes={"key": "value1"})

    attr = project.customattributes.set("key", "value2")
    attr = project.customattributes.get("key")
    assert attr.value == "value2"
    assert attr in project.customattributes.list()

    attr.delete()


def test_project_environments(project):
    environment = project.environments.create(
        {"name": "env1", "external_url": "http://fake.env/whatever"}
    )
    environments = project.environments.list()
    assert environment in environments

    environment = environments[0]
    environment.external_url = "http://new.env/whatever"
    environment.save()

    environment = project.environments.list()[0]
    assert environment.external_url == "http://new.env/whatever"

    environment.stop()

    environment.delete()


def test_project_events(project):
    events = project.events.list()
    assert isinstance(events, list)


def test_project_file_uploads(project):
    filename = "test.txt"
    file_contents = "testing contents"

    uploaded_file = project.upload(filename, file_contents)
    alt, url = uploaded_file["alt"], uploaded_file["url"]
    assert alt == filename
    assert url.startswith("/uploads/")
    assert url.endswith(f"/{filename}")
    assert uploaded_file["markdown"] == f"[{alt}]({url})"


def test_project_forks(gl, project, user):
    fork = project.forks.create({"namespace": user.username})
    fork_project = gl.projects.get(fork.id)
    assert fork_project.forked_from_project["id"] == project.id

    forks = project.forks.list()
    assert fork.id in [fork_project.id for fork_project in forks]


def test_project_hooks(project):
    hook = project.hooks.create({"url": "http://hook.url"})
    assert hook in project.hooks.list()

    hook.note_events = True
    hook.save()

    hook = project.hooks.get(hook.id)
    assert hook.note_events is True

    hook.delete()


def test_project_housekeeping(project):
    project.housekeeping()


def test_project_labels(project):
    label = project.labels.create({"name": "label", "color": "#778899"})
    labels = project.labels.list()
    assert label in labels

    label = project.labels.get("label")
    assert label == labels[0]

    label.new_name = "Label:that requires:encoding"
    label.save()
    assert label.name == "Label:that requires:encoding"
    label = project.labels.get("Label:that requires:encoding")
    assert label.name == "Label:that requires:encoding"

    label.subscribe()
    assert label.subscribed is True

    label.unsubscribe()
    assert label.subscribed is False

    label.delete()


def test_project_label_promotion(gl, group):
    """
    Label promotion requires the project to be a child of a group (not in a user namespace)

    """
    _id = uuid.uuid4().hex
    data = {
        "name": f"test-project-{_id}",
        "namespace_id": group.id,
    }
    project = gl.projects.create(data)

    label_name = "promoteme"
    promoted_label = project.labels.create({"name": label_name, "color": "#112233"})
    promoted_label.promote()

    assert any(label.name == label_name for label in group.labels.list())

    group.labels.delete(label_name)


def test_project_milestones(project):
    milestone = project.milestones.create({"title": "milestone1"})
    assert milestone in project.milestones.list()

    milestone.due_date = "2020-01-01T00:00:00Z"
    milestone.save()

    milestone.state_event = "close"
    milestone.save()

    milestone = project.milestones.get(milestone.id)
    assert milestone.state == "closed"
    assert not milestone.issues()
    assert not milestone.merge_requests()


def test_project_milestone_promotion(gl, group):
    """
    Milestone promotion requires the project to be a child of a group (not in a user namespace)

    """
    _id = uuid.uuid4().hex
    data = {
        "name": f"test-project-{_id}",
        "namespace_id": group.id,
    }
    project = gl.projects.create(data)

    milestone_title = "promoteme"
    promoted_milestone = project.milestones.create({"title": milestone_title})
    promoted_milestone.promote()

    assert any(
        milestone.title == milestone_title for milestone in group.milestones.list()
    )


def test_project_pages_domains(gl, project):
    domain = project.pagesdomains.create({"domain": "foo.domain.com"})
    assert domain in project.pagesdomains.list()
    assert domain in gl.pagesdomains.list()

    domain = project.pagesdomains.get("foo.domain.com")
    assert domain.domain == "foo.domain.com"

    domain.delete()


def test_project_protected_branches(project, gitlab_version):
    # Updating a protected branch is possible from Gitlab 15.6
    # https://docs.gitlab.com/ee/api/protected_branches.html#update-a-protected-branch
    can_update_prot_branch = gitlab_version.major > 15 or (
        gitlab_version.major == 15 and gitlab_version.minor >= 6
    )

    p_b = project.protectedbranches.create(
        {
            "name": "*-stable",
            "allow_force_push": False,
        }
    )
    assert p_b.name == "*-stable"
    assert not p_b.allow_force_push
    assert p_b in project.protectedbranches.list()

    if can_update_prot_branch:
        p_b.allow_force_push = True
        p_b.save()
        # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
        time.sleep(5)

    p_b = project.protectedbranches.get("*-stable")
    if can_update_prot_branch:
        assert p_b.allow_force_push

        p_b.delete()


def test_project_remote_mirrors(project):
    mirror_url = "https://gitlab.example.com/root/mirror.git"

    mirror = project.remote_mirrors.create({"url": mirror_url})
    assert mirror.url == mirror_url

    mirror.enabled = True
    mirror.save()

    mirror = project.remote_mirrors.list()[0]
    assert isinstance(mirror, gitlab.v4.objects.ProjectRemoteMirror)
    assert mirror.url == mirror_url
    assert mirror.enabled is True

    mirror.delete()


def test_project_services(project):
    # Use 'update' to create a service as we don't have a 'create' method and
    # to add one is somewhat complicated so it hasn't been done yet.
    project.services.update("asana", api_key="foo")

    service = project.services.get("asana")
    assert service.active is True
    service.api_key = "whatever"
    service.save()

    service = project.services.get("asana")
    assert service.active is True

    service.delete()


def test_project_stars(project):
    project.star()
    assert project.star_count == 1

    project.unstar()
    assert project.star_count == 0


def test_project_storage(project):
    storage = project.storage.get()
    assert isinstance(storage, ProjectStorage)
    assert storage.repository_storage == "default"


def test_project_tags(project, project_file):
    tag = project.tags.create({"tag_name": "v1.0", "ref": "main"})
    assert tag in project.tags.list()

    tag.delete()


def test_project_triggers(project):
    trigger = project.triggers.create({"description": "trigger1"})
    assert trigger in project.triggers.list()

    trigger.delete()


def test_project_wiki(project):
    content = "Wiki page content"
    wiki = project.wikis.create({"title": "wikipage", "content": content})
    assert wiki in project.wikis.list()

    wiki = project.wikis.get(wiki.slug)
    assert wiki.content == content

    # update and delete seem broken
    wiki.content = "new content"
    wiki.save()

    wiki.delete()


def test_project_groups_list(gl, group):
    """Test listing groups of a project"""
    # Create a subgroup of our top-group, we will place our new project inside
    # this group.
    group2 = gl.groups.create(
        {"name": "group2_proj", "path": "group2_proj", "parent_id": group.id}
    )
    data = {
        "name": "test-project-tpsg",
        "namespace_id": group2.id,
    }
    project = gl.projects.create(data)

    groups = project.groups.list()
    group_ids = set([x.id for x in groups])
    assert {group.id, group2.id} == group_ids


def test_project_transfer(gl, project, group):
    assert project.namespace["path"] != group.full_path
    project.transfer(group.id)

    project = gl.projects.get(project.id)
    assert project.namespace["path"] == group.full_path

    gl.auth()
    project.transfer(gl.user.username)

    project = gl.projects.get(project.id)
    assert project.namespace["path"] == gl.user.username



import copy
import logging
import pickle
from http.client import HTTPConnection

import pytest
import requests
import responses

import gitlab
from gitlab.config import GitlabConfigMissingError, GitlabDataError
from tests.unit import helpers

localhost = "http://localhost"
token = "abc123"


@pytest.fixture
def resp_get_user():
    return {
        "method": responses.GET,
        "url": "http://localhost/api/v4/user",
        "json": {
            "id": 1,
            "username": "username",
            "web_url": "http://localhost/username",
        },
        "content_type": "application/json",
        "status": 200,
    }


@pytest.fixture
def resp_page_1():
    headers = {
        "X-Page": "1",
        "X-Next-Page": "2",
        "X-Per-Page": "1",
        "X-Total-Pages": "2",
        "X-Total": "2",
        "Link": ("<http://localhost/api/v4/tests?per_page=1&page=2>;" ' rel="next"'),
    }

    return {
        "method": responses.GET,
        "url": "http://localhost/api/v4/tests",
        "json": [{"a": "b"}],
        "headers": headers,
        "content_type": "application/json",
        "status": 200,
        "match": helpers.MATCH_EMPTY_QUERY_PARAMS,
    }


@pytest.fixture
def resp_page_2():
    headers = {
        "X-Page": "2",
        "X-Next-Page": "2",
        "X-Per-Page": "1",
        "X-Total-Pages": "2",
        "X-Total": "2",
    }
    params = {"per_page": "1", "page": "2"}

    return {
        "method": responses.GET,
        "url": "http://localhost/api/v4/tests",
        "json": [{"c": "d"}],
        "headers": headers,
        "content_type": "application/json",
        "status": 200,
        "match": [responses.matchers.query_param_matcher(params)],
    }


def test_gitlab_init_with_valid_api_version():
    gl = gitlab.Gitlab(api_version="4")
    assert gl.api_version == "4"


def test_gitlab_init_with_invalid_api_version():
    with pytest.raises(ModuleNotFoundError, match="gitlab.v1.objects"):
        gitlab.Gitlab(api_version="1")


def test_gitlab_as_context_manager():
    with gitlab.Gitlab() as gl:
        assert isinstance(gl, gitlab.Gitlab)


def test_gitlab_enable_debug(gl):
    gl.enable_debug()

    logger = logging.getLogger("requests.packages.urllib3")
    assert logger.level == logging.DEBUG
    assert HTTPConnection.debuglevel == 1


@responses.activate
@pytest.mark.parametrize(
    "status_code,response_json,expected",
    [
        (200, {"version": "0.0.0-pre", "revision": "abcdef"}, ("0.0.0-pre", "abcdef")),
        (200, None, ("unknown", "unknown")),
        (401, None, ("unknown", "unknown")),
    ],
)
def test_gitlab_get_version(gl, status_code, response_json, expected):
    responses.add(
        method=responses.GET,
        url="http://localhost/api/v4/version",
        json=response_json,
        status=status_code,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    version = gl.version()
    assert version == expected


@responses.activate
@pytest.mark.parametrize(
    "response_json,expected",
    [
        ({"id": "1", "plan": "premium"}, {"id": "1", "plan": "premium"}),
        (None, {}),
    ],
)
def test_gitlab_get_license(gl, response_json, expected):
    responses.add(
        method=responses.GET,
        url="http://localhost/api/v4/license",
        json=response_json,
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    gitlab_license = gl.get_license()
    assert gitlab_license == expected


@responses.activate
def test_gitlab_set_license(gl):
    responses.add(
        method=responses.POST,
        url="http://localhost/api/v4/license",
        json={"id": 1, "plan": "premium"},
        status=201,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    gitlab_license = gl.set_license("yJkYXRhIjoiMHM5Q")
    assert gitlab_license["plan"] == "premium"


@responses.activate
def test_gitlab_build_list(gl, resp_page_1, resp_page_2):
    responses.add(**resp_page_1)
    obj = gl.http_list("/tests", iterator=True)
    assert len(obj) == 2
    assert obj._next_url == "http://localhost/api/v4/tests?per_page=1&page=2"
    assert obj.current_page == 1
    assert obj.prev_page is None
    assert obj.next_page == 2
    assert obj.per_page == 1
    assert obj.total_pages == 2
    assert obj.total == 2

    responses.add(**resp_page_2)
    test_list = list(obj)
    assert len(test_list) == 2
    assert test_list[0]["a"] == "b"
    assert test_list[1]["c"] == "d"


def _strip_pagination_headers(response):
    """
    https://docs.gitlab.com/ee/user/gitlab_com/index.html#pagination-response-headers
    """
    stripped = copy.deepcopy(response)

    del stripped["headers"]["X-Total-Pages"]
    del stripped["headers"]["X-Total"]

    return stripped


@responses.activate
def test_gitlab_build_list_missing_headers(gl, resp_page_1, resp_page_2):
    stripped_page_1 = _strip_pagination_headers(resp_page_1)
    stripped_page_2 = _strip_pagination_headers(resp_page_2)

    responses.add(**stripped_page_1)
    obj = gl.http_list("/tests", iterator=True)
    assert len(obj) == 0  # Lazy generator has no knowledge of total items
    assert obj.total_pages is None
    assert obj.total is None

    responses.add(**stripped_page_2)
    test_list = list(obj)
    assert len(test_list) == 2  # List has total items after making the API calls


@responses.activate
def test_gitlab_get_all_omitted_when_iterator(gl, resp_page_1, resp_page_2):
    responses.add(**resp_page_1)
    responses.add(**resp_page_2)
    result = gl.http_list("/tests", iterator=True, get_all=True)
    assert isinstance(result, gitlab.GitlabList)


def test_gitlab_strip_base_url(gl_trailing):
    assert gl_trailing.url == "http://localhost"


def test_gitlab_strip_api_url(gl_trailing):
    assert gl_trailing.api_url == "http://localhost/api/v4"


def test_gitlab_build_url(gl_trailing):
    r = gl_trailing._build_url("/projects")
    assert r == "http://localhost/api/v4/projects"


def test_gitlab_pickability(gl):
    original_gl_objects = gl._objects
    pickled = pickle.dumps(gl)
    unpickled = pickle.loads(pickled)
    assert isinstance(unpickled, gitlab.Gitlab)
    assert hasattr(unpickled, "_objects")
    assert unpickled._objects == original_gl_objects


@responses.activate
def test_gitlab_token_auth(gl, resp_get_user):
    responses.add(**resp_get_user)
    gl.auth()
    assert gl.user.username == "username"
    assert gl.user.id == 1
    assert isinstance(gl.user, gitlab.v4.objects.CurrentUser)


@responses.activate
def test_gitlab_auth_with_mismatching_url_warns():
    responses.add(
        method=responses.GET,
        url="http://first.example.com/api/v4/user",
        json={
            "username": "test-user",
            "web_url": "http://second.example.com/test-user",
        },
        content_type="application/json",
        status=200,
    )
    gl = gitlab.Gitlab("http://first.example.com")

    with pytest.warns(UserWarning):
        gl.auth()


def test_gitlab_default_url():
    gl = gitlab.Gitlab()
    assert gl.url == gitlab.const.DEFAULT_URL


@pytest.mark.parametrize(
    "args, kwargs, expected_url, expected_private_token, expected_oauth_token",
    [
        ([], {}, gitlab.const.DEFAULT_URL, None, None),
        ([None, token], {}, gitlab.const.DEFAULT_URL, token, None),
        ([localhost], {}, localhost, None, None),
        ([localhost, token], {}, localhost, token, None),
        ([localhost, None, token], {}, localhost, None, token),
        ([], {"private_token": token}, gitlab.const.DEFAULT_URL, token, None),
        ([], {"oauth_token": token}, gitlab.const.DEFAULT_URL, None, token),
        ([], {"url": localhost}, localhost, None, None),
        ([], {"url": localhost, "private_token": token}, localhost, token, None),
        ([], {"url": localhost, "oauth_token": token}, localhost, None, token),
    ],
    ids=[
        "no_args",
        "args_private_token",
        "args_url",
        "args_url_private_token",
        "args_url_oauth_token",
        "kwargs_private_token",
        "kwargs_oauth_token",
        "kwargs_url",
        "kwargs_url_private_token",
        "kwargs_url_oauth_token",
    ],
)
def test_gitlab_args_kwargs(
    args, kwargs, expected_url, expected_private_token, expected_oauth_token
):
    gl = gitlab.Gitlab(*args, **kwargs)
    assert gl.url == expected_url
    assert gl.private_token == expected_private_token
    assert gl.oauth_token == expected_oauth_token


def test_gitlab_from_config(default_config):
    config_path = default_config
    gitlab.Gitlab.from_config("one", [config_path])


def test_gitlab_from_config_without_files_raises():
    with pytest.raises(GitlabConfigMissingError, match="non-existing"):
        gitlab.Gitlab.from_config("non-existing")


def test_gitlab_from_config_with_wrong_gitlab_id_raises(default_config):
    with pytest.raises(GitlabDataError, match="non-existing"):
        gitlab.Gitlab.from_config("non-existing", [default_config])


def test_gitlab_subclass_from_config(default_config):
    class MyGitlab(gitlab.Gitlab):
        pass

    config_path = default_config
    gl = MyGitlab.from_config("one", [config_path])
    assert isinstance(gl, MyGitlab)


@pytest.mark.parametrize(
    "kwargs,expected_agent",
    [
        ({}, gitlab.const.USER_AGENT),
        ({"user_agent": "my-package/1.0.0"}, "my-package/1.0.0"),
    ],
)
def test_gitlab_user_agent(kwargs, expected_agent):
    gl = gitlab.Gitlab("http://localhost", **kwargs)
    assert gl.headers["User-Agent"] == expected_agent


def test_gitlab_enum_const_does_not_warn(recwarn):
    no_access = gitlab.const.AccessLevel.NO_ACCESS

    assert not recwarn
    assert no_access == 0


def test_gitlab_plain_const_does_not_warn(recwarn):
    no_access = gitlab.const.NO_ACCESS

    assert not recwarn
    assert no_access == 0


@responses.activate
@pytest.mark.parametrize(
    "kwargs,link_header,expected_next_url,show_warning",
    [
        (
            {},
            "<http://localhost/api/v4/tests?per_page=1&page=2>;" ' rel="next"',
            "http://localhost/api/v4/tests?per_page=1&page=2",
            False,
        ),
        (
            {},
            "<http://orig_host/api/v4/tests?per_page=1&page=2>;" ' rel="next"',
            "http://orig_host/api/v4/tests?per_page=1&page=2",
            True,
        ),
        (
            {"keep_base_url": True},
            "<http://orig_host/api/v4/tests?per_page=1&page=2>;" ' rel="next"',
            "http://localhost/api/v4/tests?per_page=1&page=2",
            False,
        ),
    ],
    ids=["url-match-does-not-warn", "url-mismatch-warns", "url-mismatch-keeps-url"],
)
def test_gitlab_keep_base_url(kwargs, link_header, expected_next_url, show_warning):
    responses.add(
        **{
            "method": responses.GET,
            "url": "http://localhost/api/v4/tests",
            "json": [{"a": "b"}],
            "headers": {
                "X-Page": "1",
                "X-Next-Page": "2",
                "X-Per-Page": "1",
                "X-Total-Pages": "2",
                "X-Total": "2",
                "Link": (link_header),
            },
            "content_type": "application/json",
            "status": 200,
            "match": helpers.MATCH_EMPTY_QUERY_PARAMS,
        }
    )

    gl = gitlab.Gitlab(url="http://localhost", **kwargs)
    if show_warning:
        with pytest.warns(UserWarning) as warn_record:
            obj = gl.http_list("/tests", iterator=True)
        assert len(warn_record) == 1
    else:
        obj = gl.http_list("/tests", iterator=True)
    assert obj._next_url == expected_next_url


def test_no_custom_session(default_config):
    """Test no custom session"""

    config_path = default_config
    custom_session = requests.Session()
    test_gitlab = gitlab.Gitlab.from_config("one", [config_path])

    assert test_gitlab.session != custom_session


def test_custom_session(default_config):
    """Test custom session"""

    config_path = default_config
    custom_session = requests.Session()
    test_gitlab = gitlab.Gitlab.from_config(
        "one", [config_path], session=custom_session
    )

    assert test_gitlab.session == custom_session



"""
GitLab API: https://docs.gitlab.com/ee/api/resource_iteration_events.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectIssueResourceIterationEvent

issue_event_content = {"id": 1, "resource_type": "Issue"}


@pytest.fixture()
def resp_list_project_issue_iteration_events():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues/1/resource_iteration_events",
            json=[issue_event_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_get_project_issue_iteration_event():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues/1/resource_iteration_events/1",
            json=issue_event_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_issue_iteration_events(
    project_issue, resp_list_project_issue_iteration_events
):
    iteration_events = project_issue.resource_iteration_events.list()
    assert isinstance(iteration_events, list)

    iteration_event = iteration_events[0]
    assert isinstance(iteration_event, ProjectIssueResourceIterationEvent)
    assert iteration_event.resource_type == "Issue"


def test_get_project_issue_iteration_event(
    project_issue, resp_get_project_issue_iteration_event
):
    iteration_event = project_issue.resource_iteration_events.get(1)
    assert isinstance(iteration_event, ProjectIssueResourceIterationEvent)
    assert iteration_event.resource_type == "Issue"



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectTrigger",
    "ProjectTriggerManager",
]


class ProjectTrigger(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectTriggerManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/triggers"
    _obj_cls = ProjectTrigger
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("description",))
    _update_attrs = RequiredOptional(required=("description",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectTrigger:
        return cast(ProjectTrigger, super().get(id=id, lazy=lazy, **kwargs))



"""
GitLab API: https://docs.gitlab.com/ce/api/bulk_imports.html
"""

import pytest
import responses

from gitlab.v4.objects import BulkImport, BulkImportAllEntity, BulkImportEntity

migration_content = {
    "id": 1,
    "status": "finished",
    "source_type": "gitlab",
    "created_at": "2021-06-18T09:45:55.358Z",
    "updated_at": "2021-06-18T09:46:27.003Z",
}
entity_content = {
    "id": 1,
    "bulk_import_id": 1,
    "status": "finished",
    "source_full_path": "source_group",
    "destination_slug": "destination_slug",
    "destination_namespace": "destination_path",
    "parent_id": None,
    "namespace_id": 1,
    "project_id": None,
    "created_at": "2021-06-18T09:47:37.390Z",
    "updated_at": "2021-06-18T09:47:51.867Z",
    "failures": [],
}


@pytest.fixture
def resp_create_bulk_import():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/bulk_imports",
            json=migration_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_list_bulk_imports():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/bulk_imports",
            json=[migration_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_bulk_import():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/bulk_imports/1",
            json=migration_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_all_bulk_import_entities():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/bulk_imports/entities",
            json=[entity_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_bulk_import_entities():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/bulk_imports/1/entities",
            json=[entity_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_bulk_import_entity():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/bulk_imports/1/entities/1",
            json=entity_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_create_bulk_import(gl, resp_create_bulk_import):
    configuration = {
        "url": gl.url,
        "access_token": "test-token",
    }
    migration_entity = {
        "source_full_path": "source",
        "source_type": "group_entity",
        "destination_slug": "destination",
        "destination_namespace": "destination",
    }
    migration = gl.bulk_imports.create(
        {
            "configuration": configuration,
            "entities": [migration_entity],
        }
    )
    assert isinstance(migration, BulkImport)
    assert migration.status == "finished"


def test_list_bulk_imports(gl, resp_list_bulk_imports):
    migrations = gl.bulk_imports.list()
    assert isinstance(migrations[0], BulkImport)
    assert migrations[0].status == "finished"


def test_get_bulk_import(gl, resp_get_bulk_import):
    migration = gl.bulk_imports.get(1)
    assert isinstance(migration, BulkImport)
    assert migration.status == "finished"


def test_list_all_bulk_import_entities(gl, resp_list_all_bulk_import_entities):
    entities = gl.bulk_import_entities.list()
    assert isinstance(entities[0], BulkImportAllEntity)
    assert entities[0].bulk_import_id == 1


def test_list_bulk_import_entities(gl, migration, resp_list_bulk_import_entities):
    entities = migration.entities.list()
    assert isinstance(entities[0], BulkImportEntity)
    assert entities[0].bulk_import_id == 1


def test_get_bulk_import_entity(gl, migration, resp_get_bulk_import_entity):
    entity = migration.entities.get(1)
    assert isinstance(entity, BulkImportEntity)
    assert entity.bulk_import_id == 1



"""
GitLab API:
https://docs.gitlab.com/ce/api/topics.html
"""

import pytest
import responses

from gitlab.v4.objects import Topic

name = "GitLab"
topic_title = "topic title"
new_name = "gitlab-test"
topic_content = {
    "id": 1,
    "name": name,
    "title": topic_title,
    "description": "GitLab is an open source end-to-end software development platform.",
    "total_projects_count": 1000,
    "avatar_url": "http://www.gravatar.com/avatar/a0d477b3ea21970ce6ffcbb817b0b435?s=80&d=identicon",
}
topics_url = "http://localhost/api/v4/topics"
topic_url = f"{topics_url}/1"


@pytest.fixture
def resp_list_topics():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=topics_url,
            json=[topic_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_topic():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=topic_url,
            json=topic_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_topic():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=topics_url,
            json=topic_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_update_topic():
    updated_content = dict(topic_content)
    updated_content["name"] = new_name

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=topic_url,
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_topic():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url=topic_url,
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_merge_topics():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=f"{topics_url}/merge",
            json=topic_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_topics(gl, resp_list_topics):
    topics = gl.topics.list()
    assert isinstance(topics, list)
    assert isinstance(topics[0], Topic)
    assert topics[0].name == name


def test_get_topic(gl, resp_get_topic):
    topic = gl.topics.get(1)
    assert isinstance(topic, Topic)
    assert topic.name == name


def test_create_topic(gl, resp_create_topic):
    topic = gl.topics.create({"name": name, "title": topic_title})
    assert isinstance(topic, Topic)
    assert topic.name == name
    assert topic.title == topic_title


def test_update_topic(gl, resp_update_topic):
    topic = gl.topics.get(1, lazy=True)
    topic.name = new_name
    topic.save()
    assert topic.name == new_name


def test_delete_topic(gl, resp_delete_topic):
    topic = gl.topics.get(1, lazy=True)
    topic.delete()


def test_merge_topic(gl, resp_merge_topics):
    topic = gl.topics.merge(123, 1)
    assert topic["id"] == 1



"""
GitLab API:
https://docs.gitlab.com/ee/api/merge_trains.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectMergeTrain

mr_content = {
    "id": 110,
    "merge_request": {
        "id": 1,
        "iid": 1,
        "project_id": 3,
        "title": "Test merge train",
        "description": "",
        "state": "merged",
        "created_at": "2020-02-06T08:39:14.883Z",
        "updated_at": "2020-02-06T08:40:57.038Z",
        "web_url": "http://gitlab.example.com/root/merge-train-race-condition/-/merge_requests/1",
    },
    "user": {
        "id": 1,
        "name": "Administrator",
        "username": "root",
        "state": "active",
        "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        "web_url": "http://gitlab.example.com/root",
    },
    "pipeline": {
        "id": 246,
        "sha": "bcc17a8ffd51be1afe45605e714085df28b80b13",
        "ref": "refs/merge-requests/1/train",
        "status": "success",
        "created_at": "2020-02-06T08:40:42.410Z",
        "updated_at": "2020-02-06T08:40:46.912Z",
        "web_url": "http://gitlab.example.com/root/merge-train-race-condition/pipelines/246",
    },
    "created_at": "2020-02-06T08:39:47.217Z",
    "updated_at": "2020-02-06T08:40:57.720Z",
    "target_branch": "feature-1580973432",
    "status": "merged",
    "merged_at": "2020-02-06T08:40:57.719Z",
    "duration": 70,
}


@pytest.fixture
def resp_list_merge_trains():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_trains",
            json=[mr_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_merge_requests(project, resp_list_merge_trains):
    merge_trains = project.merge_trains.list()
    assert isinstance(merge_trains[0], ProjectMergeTrain)
    assert merge_trains[0].id == mr_content["id"]
